"""Base class of all reinforced cross-sections."""

from abc import ABC
from collections.abc import Callable
from functools import partial

from shapely import LineString

from blueprints.materials.concrete import ConcreteMaterial
from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.structural_sections._cross_section import CrossSection
from blueprints.structural_sections.concrete.rebar import Rebar
from blueprints.structural_sections.concrete.reinforced_concrete_sections.reinforcement_configurations import (
    ReinforcementConfiguration,
)
from blueprints.structural_sections.concrete.stirrups import StirrupConfiguration
from blueprints.type_alias import KG_M, KG_M3, M3_M, MM2_M
from blueprints.unit_conversion import MM3_TO_M3


class ReinforcedCrossSection(ABC):
    """Base class of all reinforced cross-sections."""

    def __init__(
        self,
        cross_section: CrossSection,
        concrete_material: ConcreteMaterial,
    ) -> None:
        """Initialize the reinforced cross-section.

        Parameters
        ----------
        cross_section : CrossSection
            Cross-section of the reinforced concrete section.
        concrete_material : ConcreteMaterial
            Material properties of the concrete.
        """
        self.cross_section = cross_section
        self.concrete_material = concrete_material
        self._reinforcement_configurations: list[tuple[LineString | Callable[..., LineString], ReinforcementConfiguration]] = []
        self._single_longitudinal_rebars: list[Rebar] = []
        self._stirrups: list[StirrupConfiguration] = []

    @property
    def longitudinal_rebars(self) -> list[Rebar]:
        """Return a list of all longitudinal rebars."""
        rebars: list[Rebar] = []

        # add the single longitudinal rebars
        rebars.extend(self._single_longitudinal_rebars)

        # add the rebars from the reinforcement configurations
        for line, configuration in self._reinforcement_configurations:
            if callable(line):
                # partial function with additional arguments where the line should be called to get the LineString
                # the implementation will be made at that level and inserted here to produce the rebars needed.
                # this keeps this ABC class clean and allows for a lot of flexibility in the implementation of the line.
                # this has been done to be able to add any shape of line to the cross-section (e.g. a circle or any other in the future).
                rebars.extend(configuration.to_rebars(line=line()))
            else:
                rebars.extend(configuration.to_rebars(line=line))

        # check if all rebars are inside the cross-section.
        # needed for the case where custom configurations are added to the RCS
        for rebar in rebars:
            if not self.cross_section.polygon.contains(other=rebar.polygon):
                msg = f"Rebar (diameter={rebar.diameter}, x={rebar.x}, y={rebar.y}) is not (fully) inside the cross-section."
                raise ValueError(msg)

        return rebars

    @property
    def stirrups(self) -> list[StirrupConfiguration]:
        """Return a list of all stirrups."""
        return self._stirrups

    @property
    def reinforcement_weight_longitudinal_bars(self) -> KG_M:
        """Total mass of the longitudinal reinforcement in the cross-section per meter length [kg/m]."""
        return sum(rebar.weight_per_meter for rebar in self.longitudinal_rebars)

    @property
    def reinforcement_weight_stirrups(self) -> KG_M:
        """Total mass of the stirrups' reinforcement in the cross-section per meter length [kg/m]."""
        return sum(stirrup.weight_per_meter for stirrup in self._stirrups)

    @property
    def reinforcement_weight(self) -> KG_M:
        """Total mass of the reinforcement in the cross-section per meter length [kg/m]."""
        return self.reinforcement_weight_longitudinal_bars + self.reinforcement_weight_stirrups

    @property
    def reinforcement_area_longitudinal_bars(self) -> MM2_M:
        """Total area of the longitudinal reinforcement in the cross-section per meter length [mm²/m]."""
        return sum(rebar.area for rebar in self.longitudinal_rebars)

    @property
    def concrete_volume(self) -> M3_M:
        """Total volume of the reinforced cross-section per meter length [m³/m]."""
        length = 1000  # mm
        return self.cross_section.area * length * MM3_TO_M3

    @property
    def weight_per_volume(self) -> KG_M3:
        """Total mass of the cross-section per meter length (concrete+reinforcement) [kg/m³]."""
        return self.reinforcement_weight / self.concrete_volume

    def get_present_steel_materials(self) -> list[ReinforcementSteelMaterial]:
        """Return a list of all present steel materials in the cross-section."""
        materials = [rebar.material for rebar in self.longitudinal_rebars]
        materials.extend(stirrup.material for stirrup in self._stirrups)
        return list(set(materials))

    def add_longitudinal_rebar(
        self,
        rebar: Rebar,
    ) -> Rebar:
        """Adds a single reinforcement bar to the cross-section.

        Parameters
        ----------
        rebar : Rebar
            Rebar to be added to the cross-section.

        Raises
        ------
        ValueError
            If the rebar is not fully inside the cross-section.

        Returns
        -------
        Rebar
            Newly created Rebar
        """
        # check if given diameter/coordinates are fully inside the cross-section
        if not rebar.polygon.within(self.cross_section.polygon):
            msg = f"Rebar (diameter={rebar.diameter}, x={rebar.x}, y={rebar.y}) is not (fully) inside the cross-section."
            raise ValueError(msg)

        # add the rebar to the list of longitudinal rebars
        self._single_longitudinal_rebars.append(rebar)

        return rebar

    def add_stirrup_configuration(self, stirrup: StirrupConfiguration) -> StirrupConfiguration:
        """Add a stirrup configuration to the cross-section.

        Parameters
        ----------
        stirrup : StirrupConfiguration
            Configuration of stirrup reinforcement in the cross-section.

        Returns
        -------
        StirrupConfiguration
            Newly created Stirrup

        Raises
        ------
        ValueError
            If the stirrup is not fully inside the cross-section.
        """
        # check if the stirrup is inside the cross-section
        stirrup_outside_edge = stirrup.geometry.buffer(distance=stirrup.diameter / 2)
        if not self.cross_section.polygon.contains(stirrup_outside_edge):
            msg = "Stirrup is not (fully) inside the cross-section."
            raise ValueError(msg)

        # add the stirrup to the list
        self._stirrups.append(stirrup)

        return stirrup

    def add_reinforcement_configuration(
        self,
        line: LineString | Callable[..., LineString],
        configuration: ReinforcementConfiguration,
        *args,
        **kwargs,
    ) -> None:
        """Add a reinforcement configuration to the cross-section.

        Parameters
        ----------
        line : LineString | Callable[..., LineString]
            Representing the path of the reinforcement in the cross-section.
            Start of the line defines the first rebar of the configuration, end of the line defines the last rebar.
            If a callable is given, it should return a LineString. The callable can take additional arguments.
            Arguments can be passed to the callable using the *args and **kwargs.
        configuration : ReinforcementConfiguration
            Configuration of the reinforcement.
        args : Any
            Additional arguments for the callable line. If line is not a callable, these arguments are ignored.
        kwargs : Any
            Additional keyword arguments for the callable line. If line is not a callable, these arguments are ignored.

        """
        # check if the line is a callable and wrap it with the given arguments
        if callable(line):
            line = partial(line, *args, **kwargs)  # type: ignore[misc]

        # add the reinforcement configuration to the list
        self._reinforcement_configurations.append((line, configuration))



"""Testing formula A.1 from EN 1993-1-9:2005: Annex A - Determination of fatigue load parameters and verification formats."""

import pytest

from blueprints.codes.eurocode.en_1993_1_9_2005.annex_a_determination_of_fatigue_load_parameters_and_verification_formats.formula_a_1 import (
    FormADot1DamageDuringDesignLife,
)
from blueprints.validations import LessOrEqualToZeroError, ListsNotSameLengthError, NegativeValueError


class TestFormADot1DamageDuringDesignLife:
    """Validation for formula A.1 from EN 1993-1-9:2005."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        n_e = [5.0, 4.0, 3.0]  # [-]
        n_r = [10.0, 20.0, 30.0]  # [-]

        form_a_1 = FormADot1DamageDuringDesignLife(n_e=n_e, n_r=n_r)
        # manually calculated result
        manually_calculated_result = 5 / 10 + 4 / 20 + 3 / 30  # -

        assert form_a_1 == pytest.approx(expected=manually_calculated_result, rel=1e-9)

    def test_raise_error_if_negative_n_e(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for n_e."""
        n_e = [-5.0, 4.0, 3.0]  # [-]
        n_r = [10.0, 20.0, 30.0]  # [-]
        with pytest.raises(NegativeValueError):
            FormADot1DamageDuringDesignLife(n_e=n_e, n_r=n_r)

    @pytest.mark.parametrize(
        ("n_e", "n_r"),
        [
            ([5.0, 4.0, 3.0], [10.0, -20.0, 30.0]),
            ([5.0, 4.0, 3.0], [10.0, 20.0, 0.0]),
        ],
    )
    def test_raise_error_if_negative_n_r(self, n_e: list[float], n_r: list[float]) -> None:
        """Test that a LessOrEqualToZeroError is raised when a negative value is passed for n_r."""
        with pytest.raises(LessOrEqualToZeroError):
            FormADot1DamageDuringDesignLife(n_e=n_e, n_r=n_r)

    def test_raise_error_if_n_e_and_n_r_are_different_length(self) -> None:
        """Test that a ListsNotSameLengthError is raised when n_e and n_r are not the same length."""
        n_e = [5.0, 4.0]  # [-]
        n_r = [10.0, 20.0, 30.0]  # [-]
        with pytest.raises(ListsNotSameLengthError):
            FormADot1DamageDuringDesignLife(n_e=n_e, n_r=n_r)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"D_d = \sum_{i}^{n} \frac{n_{Ei}}{N_Ri} = \frac{5.000}{10.000} + \frac{4.000}{20.000} + \frac{3.000}{30.000} = 0.800",
            ),
            ("short", r"D_d = 0.800"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_e = [5.0, 4.0, 3.0]  # [-]
        n_r = [10.0, 20.0, 30.0]  # [-]

        # Object to test
        form_a_1_latex = FormADot1DamageDuringDesignLife(n_e=n_e, n_r=n_r).latex()

        actual = {
            "complete": form_a_1_latex.complete,
            "short": form_a_1_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 5.8 of EN 1993-5:2007."""

import pytest

from blueprints.codes.eurocode.en_1993_5_2007.chapter_5_ultimate_limit_states.formula_5_8 import Form5Dot8RelativeWebSlenderness
from blueprints.validations import LessOrEqualToZeroError


class TestForm5Dot8RelativeWebSlenderness:
    """Validation for formula 5.8 from EN 1993-5:2007."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c = 500  # MM
        t_w = 10  # MM
        f_y = 250  # MPA
        e = 200000  # MPA

        form = Form5Dot8RelativeWebSlenderness(c=c, t_w=t_w, f_y=f_y, e=e)

        # Expected result, manually calculated
        expected = 0.611647366

        assert form == pytest.approx(expected)

    @pytest.mark.parametrize(
        ("c", "t_w", "f_y", "e"),
        [
            (0, 10, 250, 200000),  # c is zero
            (-500, 10, 250, 200000),  # c is negative
            (500, 0, 250, 200000),  # t_w is zero
            (500, -10, 250, 200000),  # t_w is negative
            (500, 10, 0, 200000),  # f_y is zero
            (500, 10, -250, 200000),  # f_y is negative
            (500, 10, 250, 0),  # E is zero
            (500, 10, 250, -200000),  # E is negative
        ],
    )
    def test_raise_error_when_negative_or_zero_values_are_given(self, c: float, t_w: float, f_y: float, e: float) -> None:
        """Test a zero and negative value for parameters c, t_w, f_y, and e."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot8RelativeWebSlenderness(c=c, t_w=t_w, f_y=f_y, e=e)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\overline{\lambda} = 0.346 \cdot \frac{c}{t_w} \sqrt{\frac{f_y}{E}} = "
                r"0.346 \cdot \frac{500}{10} \sqrt{\frac{250}{200000}} = 0.61",
            ),
            ("short", r"\overline{\lambda} = 0.61"),
            (
                "string",
                r"\overline{\lambda} = 0.346 \cdot \frac{c}{t_w} \sqrt{\frac{f_y}{E}} = "
                r"0.346 \cdot \frac{500}{10} \sqrt{\frac{250}{200000}} = 0.61",
            ),
        ],
    )
    def test_latex_output(self, representation: str, expected: str) -> None:
        """Test the latex implementation."""
        c = 500  # MM
        t_w = 10  # MM
        f_y = 250  # MPA
        e = 200000  # MPA

        form = Form5Dot8RelativeWebSlenderness(c=c, t_w=t_w, f_y=f_y, e=e).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Generate the index page using the README.md file and fixing the cross links."""

import re
from pathlib import Path

import mkdocs_gen_files

root = Path(__file__).parents[2]
readme_file = root / "README.md"


def replace_links(readme_content: str) -> str:
    """Replace the relative links to the overview page by removing the .md extension."""
    readme_content = readme_content.replace("docs/", "")
    # replace the relative links to the overview page by removing the .md extension
    readme_content = re.sub(r"/\w+\.md", lambda match: match.group(0)[:-3], readme_content)
    # replace blueprints/code references by links to the API reference
    readme_content = readme_content.replace('"blueprints/codes', '"API reference/codes')
    # add 'https://github.com/Blueprints-org/blueprints' prefix to file in [link](file) if file is not a URL
    return re.sub(r"\[([^\]]+)\]\((?!http)([^)]+)\)", r"[\1](https://github.com/Blueprints-org/blueprints/blob/main/\2)", readme_content)


with open(readme_file, encoding="utf-8") as fd:
    readme_content = fd.read()
    readme_content = replace_links(readme_content)
    # remove the read the docs chapter, we're already in the docs
    readme_content = re.sub(r"## Read the docs!\n\n.*\n## Quick", "## Quick", readme_content, flags=re.DOTALL)

with mkdocs_gen_files.open("index.md", "w") as fd:
    # hide navigation in the index page
    fd.writelines(["---\n", "hide:\n", "  - navigation\n", "---\n"])
    fd.write(readme_content)



"""Testing formula 5.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_4 import Form5Dot4TransverseForceEffectBracingSystem
from blueprints.validations import NegativeValueError


class TestForm5Dot4TransverseForceEffectBracingSystem:
    """Validation for formula 5.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        theta_i = 0.003  # -
        n_a = 5  # kN
        n_b = 10  # kN

        # Object to test
        form_5_4 = Form5Dot4TransverseForceEffectBracingSystem(theta_i=theta_i, n_a=n_a, n_b=n_b)

        # Expected result, manually calculated
        manually_calculated_result = 0.015  # kN

        assert form_5_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("theta_i", "n_a", "n_b"),
        [
            (-0.003, 5, 10),
            (0.003, -5, 10),
            (0.003, 5, -10),
        ],
    )
    def test_raise_error_when_negative_theta_i_is_given(
        self,
        theta_i: float,
        n_a: float,
        n_b: float,
    ) -> None:
        """Test negative values for theta_i, n_a and n_b."""
        with pytest.raises(NegativeValueError):
            Form5Dot4TransverseForceEffectBracingSystem(theta_i=theta_i, n_a=n_a, n_b=n_b)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"H_{i} = Θ_{i} \cdot (N_{b} - N_{a}) = 0.003 \cdot (10.000 - 5.000) = 0.015",
            ),
            ("short", r"H_{i} = 0.015"),
            (
                "string",
                r"H_{i} = Θ_{i} \cdot (N_{b} - N_{a}) = 0.003 \cdot (10.000 - 5.000) = 0.015",
            ),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        theta_i = 0.003  # -
        n_a = 5  # kN
        n_b = 10  # kN

        # Object to test
        form_5_4_latex = Form5Dot4TransverseForceEffectBracingSystem(
            theta_i=theta_i,
            n_a=n_a,
            n_b=n_b,
        ).latex()

        actual = {
            "complete": form_5_4_latex.complete,
            "short": form_5_4_latex.short,
            "string": str(form_5_4_latex),
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 3.24 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_24_25 import (
    Form3Dot24And25IncreasedCharacteristicCompressiveStrength,
)
from blueprints.validations import NegativeValueError


class TestForm3Dot24And25IncreasedCharacteristicCompressiveStrength:
    """Validation for formula 3.24 and 3.25 from EN 1992-1-1:2004."""

    def test_evaluation_small_sigma_2(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 0.03 * f_ck  # MPa

        form_3_24 = Form3Dot24And25IncreasedCharacteristicCompressiveStrength(f_ck=f_ck, sigma_2=sigma_2)

        # Expected result, manually calculated
        manually_calculated_result = 14.03

        assert form_3_24 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_ck_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        f_ck = -12.2  # MPa
        sigma_2 = 0.03 * f_ck  # MPa

        with pytest.raises(NegativeValueError):
            Form3Dot24And25IncreasedCharacteristicCompressiveStrength(f_ck=f_ck, sigma_2=sigma_2)

    def test_evaluation_large_sigma_2(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 0.08 * f_ck  # MPa

        form_3_25 = Form3Dot24And25IncreasedCharacteristicCompressiveStrength(f_ck=f_ck, sigma_2=sigma_2)

        # Expected result, manually calculated
        manually_calculated_result = 16.165

        assert form_3_25 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ck,c} = f_{ck} \cdot (1.000 + 5.0 \cdot \sigma_2 / f_{ck}) = 12.20 \cdot (1.000 + 5.0 \cdot 0.50 / 12.20) = 14.70",
            ),
            ("short", r"f_{ck,c} = 14.70"),
        ],
    )
    def test_latex_below_0dot05fck(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 0.5  # MPa

        # Object to test
        form_3_24_25_latex = Form3Dot24And25IncreasedCharacteristicCompressiveStrength(f_ck=f_ck, sigma_2=sigma_2).latex()

        actual = {"complete": form_3_24_25_latex.complete, "short": form_3_24_25_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ck,c} = f_{ck} \cdot (1.125 + 2.5 \cdot \sigma_2 / f_{ck}) = 12.20 \cdot (1.125 + 2.5 \cdot 1.00 / 12.20) = 16.23",
            ),
            ("short", r"f_{ck,c} = 16.23"),
        ],
    )
    def test_latex_above_0dot05fck(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 1  # MPa

        # Object to test
        form_3_24_25_latex = Form3Dot24And25IncreasedCharacteristicCompressiveStrength(f_ck=f_ck, sigma_2=sigma_2).latex()

        actual = {"complete": form_3_24_25_latex.complete, "short": form_3_24_25_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test the RHS enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhs import RHS


class TestRHS:
    """Tests for the RHS enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert RHS.RHS50x30_2_6.value == ("RHS50x30x2.6", 50, 30, 2.6, 3.9, 2.6)
        assert RHS.RHS500x300_20.value == ("RHS500x300x20", 500, 300, 20, 30, 20)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "RHS50x30x2.6" in [e.value[0] for e in RHS]
        assert "RHS500x300x20" in [e.value[0] for e in RHS]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in RHS]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = RHS.RHS50x30_2_6
        assert profile.alias == "RHS50x30x2.6"
        assert profile.total_height == 50
        assert profile.total_width == 30
        assert profile.thickness == 2.6
        assert profile.outer_radius == 3.9
        assert profile.inner_radius == 2.6



"""Testing formula 3.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_14 import Form3Dot14StressStrainForShortTermLoading


class TestForm3Dot14StressStrainForShortTermLoading:
    """Validation for formula 3.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        k = 0.38  # -
        eta = 0.88  # -

        form_3_14 = Form3Dot14StressStrainForShortTermLoading(k=k, eta=eta)

        # Expected result, manually calculated
        manually_calculated_result = 1.03383

        assert form_3_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_k_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        k = -0.38  # -
        eta = 0.88  # -

        with pytest.raises(ValueError):
            Form3Dot14StressStrainForShortTermLoading(k=k, eta=eta)

    def test_raise_error_when_negative_eta_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        k = 0.38  # -
        eta = -0.88  # -

        with pytest.raises(ValueError):
            Form3Dot14StressStrainForShortTermLoading(k=k, eta=eta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\frac{\sigma_c}{f_{cm}} = \frac{k \cdot \eta - \eta^2}{1 + (k-2) \cdot \eta} = "
                r"\frac{0.380 \cdot 0.880 - 0.880^2}{1 + (0.380-2) \cdot 0.880} = 1.034",
            ),
            ("short", r"\frac{\sigma_c}{f_{cm}} = 1.034"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k = 0.38  # -
        eta = 0.88

        # Object to test
        form_3_14_latex = Form3Dot14StressStrainForShortTermLoading(k=k, eta=eta).latex()

        actual = {"complete": form_3_14_latex.complete, "short": form_3_14_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 3.7 from EN 1992-1-1:2004: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form3Dot7NonLinearCreepCoefficient(Formula):
    """Class representing formula 3.7 for the calculation of the non-linear creep coefficient."""

    label = "3.7"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        phi_inf_t0: DIMENSIONLESS,
        k_sigma: DIMENSIONLESS,
    ) -> None:
        r"""[$\varphi_{nl}(\infty,t_0)$] The non-linear creep coefficient [$-$].

        EN 1992-1-1:2004 art.3.1.4(4) - Formula (3.7)

        Parameters
        ----------
        phi_inf_t0 : DIMENSIONLESS
            [$\varphi(\infty, t_0)$] Creep coefficient if high accuracy is not required use figure 3.1 and/or use appendix B [$-$].
        k_sigma : DIMENSIONLESS
            [$k_{\sigma}$] Stress-strength ratio [$\sigma_c / f_{ck}(t_0)$] [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.phi_inf_t0 = phi_inf_t0
        self.k_sigma = k_sigma

    @staticmethod
    def _evaluate(
        phi_inf_t0: DIMENSIONLESS,
        k_sigma: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma)
        return phi_inf_t0 * np.exp(1.5 * (k_sigma - 0.45))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.7."""
        return LatexFormula(
            return_symbol=r"\varphi_{nl}(\infty, t_0)",
            result=f"{self:.{n}f}",
            equation=r"\varphi(\infty, t_0) \cdot \exp( 1.5 ( k_{\sigma} - 0.45))",
            numeric_equation=rf"{self.phi_inf_t0:.{n}f} \cdot \exp( 1.5 ( {self.k_sigma:.{n}f} - 0.45))",
            comparison_operator_label="=",
        )



"""Testing formula 6.41 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_41 import Form6Dot41W1Rectangular
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot41W1Rectangular:
    """Validation for formula 6.41 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        c_1 = 300.0
        c_2 = 400.0
        d = 500.0

        # Object to test
        formula = Form6Dot41W1Rectangular(c_1=c_1, c_2=c_2, d=d)

        # Expected result, manually calculated
        manually_calculated_result = 5907477.796  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("c_1", "c_2", "d"),
        [
            (-300.0, 400.0, 500.0),  # c_1 is negative
            (300.0, -400.0, 500.0),  # c_2 is negative
            (300.0, 400.0, -500.0),  # d is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, c_1: float, c_2: float, d: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot41W1Rectangular(c_1=c_1, c_2=c_2, d=d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"W_1 = \frac{c_1^2}{2} + c_1 \cdot c_2 + 4 \cdot c_2 \cdot d + 16 \cdot d^2 + 2 \cdot \pi \cdot d \cdot c_1 = "
                r"\frac{300.000^2}{2} + 300.000 \cdot 400.000 + 4 \cdot 400.000 \cdot 500.000 + "
                r"16 \cdot 500.000^2 + 2 \cdot \pi \cdot 500.000 \cdot 300.000 = 5907477.796 \ mm^2",
            ),
            ("short", r"W_1 = 5907477.796 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c_1 = 300.0
        c_2 = 400.0
        d = 500.0

        # Object to test
        latex = Form6Dot41W1Rectangular(c_1=c_1, c_2=c_2, d=d).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.45 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_45 import Form5Dot45LossesDueToFriction
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot45LossesDueToFriction:
    """Validation for formula 5.45 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        p_max = 1000.0
        mu = 0.2
        theta = 0.1
        k = 0.01
        x = 50.0

        # Object to test
        formula = Form5Dot45LossesDueToFriction(p_max=p_max, mu=mu, theta=theta, k=k, x=x)

        # Expected result, manually calculated
        manually_calculated_result = 113.079563283

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("p_max", "mu", "theta", "k", "x"),
        [
            (-1000.0, 0.2, 0.1, 0.01, 50.0),  # p_max is negative
            (1000.0, -0.2, 0.1, 0.01, 50.0),  # mu is negative
            (1000.0, 0.2, -0.1, 0.01, 50.0),  # theta is negative
            (1000.0, 0.2, 0.1, -0.01, 50.0),  # k is negative
            (1000.0, 0.2, 0.1, 0.01, -50.0),  # x is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, p_max: float, mu: float, theta: float, k: float, x: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot45LossesDueToFriction(p_max=p_max, mu=mu, theta=theta, k=k, x=x)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\Delta P_{\mu}(x) = P_{max} \cdot \left( 1 - e^{-\mu \cdot (\theta + k \cdot x)} \right) = "
                r"1000.000 \cdot \left( 1 - e^{-0.200 \cdot (0.100 + 0.010 \cdot 50.000)} \right) = 113.080 \ kN",
            ),
            ("short", r"\Delta P_{\mu}(x) = 113.080 \ kN"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        p_max = 1000.0
        mu = 0.2
        theta = 0.1
        k = 0.01
        x = 50.0

        # Object to test
        latex = Form5Dot45LossesDueToFriction(p_max=p_max, mu=mu, theta=theta, k=k, x=x).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 7.12 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_12 import Form7Dot12EquivalentDiameter
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot12EquivalentDiameter:
    """Validation for formula 7.12 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        n_1 = 4
        diam_1 = 16.0
        n_2 = 6
        diam_2 = 20.0

        # Object to test
        formula = Form7Dot12EquivalentDiameter(n_1=n_1, diam_1=diam_1, n_2=n_2, diam_2=diam_2)

        # Expected result, manually calculated
        manually_calculated_result = 18.609  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("n_1", "diam_1", "n_2", "diam_2"),
        [
            (-4, 16.0, 6, 20.0),  # n_1 is negative
            (4, -16.0, 6, 20.0),  # diam_1 is negative
            (4, 16.0, -6, 20.0),  # n_2 is negative
            (4, 16.0, 6, -20.0),  # diam_2 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, n_1: float, diam_1: float, n_2: float, diam_2: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot12EquivalentDiameter(n_1=n_1, diam_1=diam_1, n_2=n_2, diam_2=diam_2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"⌀_{eq} = \frac{n_1 \cdot ⌀_1^2 + n_2 \cdot ⌀_2^2}{n_1 \cdot ⌀_1 + n_2 \cdot ⌀_2} = "
                r"\frac{4.000 \cdot 16.000^2 + 6.000 \cdot 20.000^2}{4.000 \cdot 16.000 + 6.000 \cdot 20.000} = 18.609 \ mm",
            ),
            ("short", r"⌀_{eq} = 18.609 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_1 = 4
        diam_1 = 16.0
        n_2 = 6
        diam_2 = 20.0

        # Object to test
        latex = Form7Dot12EquivalentDiameter(n_1=n_1, diam_1=diam_1, n_2=n_2, diam_2=diam_2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.25 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_25 import (
    Form6Dot25CheckCombinedShearForceAndTorsionalMoment,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot25CheckCombinedShearForceAndTorsionalMoment:
    """Validation for formula 6.25 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 100.0
        v_pl_t_rd = 150.0

        # Object to test
        formula = Form6Dot25CheckCombinedShearForceAndTorsionalMoment(v_ed=v_ed, v_pl_t_rd=v_pl_t_rd)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("v_ed", "v_pl_t_rd"),
        [
            (-100.0, 150.0),  # v_ed is negative
            (100.0, -150.0),  # v_pl_t_rd is negative
            (100.0, 0.0),  # v_pl_t_rd is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, v_pl_t_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot25CheckCombinedShearForceAndTorsionalMoment(v_ed=v_ed, v_pl_t_rd=v_pl_t_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{V_{Ed}}{V_{pl,T,Rd}} \leq 1.0 \right) \to "
                r"\left( \frac{100.000}{150.000} \leq 1.0 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 100.0
        v_pl_t_rd = 150.0

        # Object to test
        latex = Form6Dot25CheckCombinedShearForceAndTorsionalMoment(v_ed=v_ed, v_pl_t_rd=v_pl_t_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 8.11 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form8Dot11MinimumDesignLapLength(Formula):
    """Class representing formula 8.11 for the calculation of the minimum design lap length [$l_{0,min}$] [$mm$]."""

    label = "8.11"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        diameter: MM,
    ) -> None:
        r"""[$l_{0,min}$] Design minimum lap length [$mm$].

        EN 1992-1-1:2004 art.8.7.3(1) - Formula (8.11)

        Parameters
        ----------
        alpha_6 : DIMENSIONLESS
            [$α_6$] Coefficient for the effect of reinforcement ratio [$-$].

            [$= (\rho_l/25)^{0.5} \leq 1.5$] with a minimum of 1.0.

            Where: [$\rho_l$] = reinforcement percentage lapped within [$0.65 \cdot l_0$] from the centre of the lap length
            considered (see figure 8.8) [$-$].

            Use your own implementation for this value or use the :class:`SubForm8Dot10Alpha6` class.
        l_b_rqd: MM
            [$l_{b,rqd}$] Basic required anchorage length, for anchoring the force [$A_s \cdot \sigma_{sd}$] in a straight bar assuming constant
            bond stress (formula 8.3) [$mm$].

            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        diameter : MM
            [$Ø$] Diameter of the bar [$mm$].
        """
        super().__init__()
        self.alpha_6 = alpha_6
        self.l_b_rqd = l_b_rqd
        self.diameter = diameter

    @staticmethod
    def _evaluate(
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        diameter: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            diameter=diameter,
        )
        return max(0.3 * alpha_6 * l_b_rqd, 15 * diameter, 200)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns a representation of the formula in LaTeX format."""
        arg_1_equation = r"0.3 \cdot \alpha_6 \cdot l_{b,rqd}"
        arg_2_equation = r"15 \cdot Ø"
        arg_3_equation = r"200 \ \text{mm}"

        arg_1_numerical_equation = rf"0.3 \cdot {self.alpha_6:.{n}f} \cdot {self.l_b_rqd:.{n}f}"
        arg_2_numerical_equation = rf"15 \cdot {self.diameter}"
        arg_3_numerical_equation = r"200"

        return LatexFormula(
            return_symbol=r"l_{Ø,min}",
            result=f"{self:.{n}f}",
            equation=f"{latex_max_curly_brackets(arg_1_equation, arg_2_equation, arg_3_equation)}",
            numeric_equation=f"{latex_max_curly_brackets(arg_1_numerical_equation, arg_2_numerical_equation, arg_3_numerical_equation)}",
            comparison_operator_label="=",
        )



"""Testing formula 6.22 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_22 import Form6Dot22CheckShearBucklingResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot22CheckShearBucklingResistance:
    """Validation for formula 6.22 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        h_w = 500.0
        t_w = 5.0
        epsilon = 1.0
        eta = 1.0

        # Object to test
        formula = Form6Dot22CheckShearBucklingResistance(h_w=h_w, t_w=t_w, epsilon=epsilon, eta=eta)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("h_w", "t_w", "epsilon", "eta"),
        [
            (-500.0, 5.0, 1.0, 1.0),  # h_w is negative
            (500.0, 5.0, -1.0, 1.0),  # epsilon is negative
            (500.0, -5.0, 1.0, 1.0),  # t_w is negative
            (500.0, 5.0, 1.0, -1.0),  # eta is negative
            (500.0, 5.0, 1.0, 0.0),  # eta is less than or equal to zero
            (500.0, 0.0, 1.0, 1.0),  # t_w is less than or equal to zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, h_w: float, t_w: float, epsilon: float, eta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot22CheckShearBucklingResistance(h_w=h_w, t_w=t_w, epsilon=epsilon, eta=eta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{h_w}{t_w} > 72 \cdot \frac{\epsilon}{\eta} \right) \to "
                r"\left( \frac{500.000}{5.000} > 72 \cdot \frac{1.000}{1.000} \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        h_w = 500.0
        t_w = 5.0
        epsilon = 1.0
        eta = 1.0

        # Object to test
        latex = Form6Dot22CheckShearBucklingResistance(h_w=h_w, t_w=t_w, epsilon=epsilon, eta=eta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 9.4 from EN 1992-1-1:2004: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DEG, DIMENSIONLESS, MM, MM2
from blueprints.validations import raise_if_greater_than_90, raise_if_negative


class Form9Dot4ShearReinforcementRatio(Formula):
    """Class representing the formula 9.4 for the calculation of the shear reinforcement ratio."""

    label = "9.4"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_sw: MM2,
        s: MM,
        b_w: MM,
        alpha: DEG,
    ) -> None:
        r"""[$\rho_w$] Shear reinforcement ratio [$-$].

        EN 1992-1-1:2004 art.9.2.2(5) - Formula (9.4)

        Parameters
        ----------
        a_sw: MM2
            [$A_{sw}$] Area of shear reinforcement within length s [$mm^2$].
        s: MM
            [$s$] The spacing between shear reinforcement along the longitudinal axis of the element [$mm$].
        b_w: MM
            [$b_w$] The width of the web of the element [$mm$].
        alpha: DEG
            [$\alpha$] The angle between the shear reinforcement and the longitudinal axis of the beam (see 9.2.2(1)) [$deg$].
        """
        super().__init__()
        self.a_sw = a_sw
        self.s = s
        self.b_w = b_w
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        a_sw: MM2,
        s: MM,
        b_w: MM,
        alpha: DEG,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(a_sw=a_sw, s=s, b_w=b_w, alpha=alpha)
        raise_if_greater_than_90(alpha=alpha)

        # Convert the angle from degrees to radians
        alpha_radians = np.deg2rad(alpha)

        return a_sw / (s * b_w * np.sin(alpha_radians))

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.4."""
        return LatexFormula(
            return_symbol=r"\rho_w",
            result=f"{self:.6f}",
            equation=r"\frac{A_{sw}}{s \cdot b_w \cdot sin(\alpha)}",
            numeric_equation=rf"\frac{{{self.a_sw:.{n}f}}}{{{self.s:.{n}f} \cdot {self.b_w:.{n}f} \cdot sin({self.alpha:.{n}f})}}",
            comparison_operator_label="=",
        )



"""Formula 6.20 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM, MM3, MM4, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot20ShearStress(Formula):
    r"""Class representing formula 6.20 for the calculation of [$\tau_{Ed}$]."""

    label = "6.20"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        v_ed: N,
        s: MM3,
        i: MM4,
        t: MM,
    ) -> None:
        r"""[$\tau_{Ed}$] Calculation of the design elastic shear stress [$MPa$].

        EN 1993-1-1:2005 art.6.2.6(4) - Formula (6.20)

        Parameters
        ----------
        v_ed : N
            [$V_{Ed}$] Design value of the shear force [$N$].
        s : MM3
            [$S$] First moment of area about the centroidal axis of that portion of the cross-section between
            the point at which the shear is required and the boundary of the cross-section [$mm^3$].
        i : MM4
            [$I$] Second moment of area of the whole cross section [$mm^4$].
        t : MM
            [$t$] Thickness at the examined point [$mm$].
        """
        super().__init__()
        self.v_ed = v_ed
        self.s = s
        self.i = i
        self.t = t

    @staticmethod
    def _evaluate(
        v_ed: N,
        s: MM3,
        i: MM4,
        t: MM,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(v_ed=v_ed, s=s)
        raise_if_less_or_equal_to_zero(i=i, t=t)

        return (v_ed * s) / (i * t)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.20."""
        _equation: str = r"\frac{V_{Ed} \cdot S}{I \cdot t}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"V_{Ed}": f"{self.v_ed:.{n}f}",
                r"S": f"{self.s:.{n}f}",
                r"I": f"{self.i:.{n}f}",
                r" t": f" {self.t:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\tau_{Ed}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Generic concrete fixtures."""

import pytest

from blueprints.materials.concrete import ConcreteMaterial, ConcreteStrengthClass
from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial, ReinforcementSteelQuality


@pytest.fixture
def fixture_concrete_material_c30_37() -> ConcreteMaterial:
    """Fixture for concrete material C30/37."""
    return ConcreteMaterial(concrete_class=ConcreteStrengthClass.C30_37)


@pytest.fixture
def fixture_concrete_material_c90_105() -> ConcreteMaterial:
    """Fixture for concrete material C90/105."""
    return ConcreteMaterial(concrete_class=ConcreteStrengthClass.C90_105)


@pytest.fixture
def fixture_reinforcement_steel_material_b500b() -> ReinforcementSteelMaterial:
    """Fixture for reinforcement steel material B500B."""
    return ReinforcementSteelMaterial(steel_quality=ReinforcementSteelQuality.B500B)



"""Testing formula 3.10 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_10 import Form3Dot10CoefficientAgeConcreteDryingShrinkage


class TestForm3Dot10CoefficientAgeConcreteDryingShrinkage:
    """Validation for formula 3.10 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        t = 10  # -
        t_s = 2  # -
        h_0 = 200  # -
        form_3_10 = Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0)

        # Expected result, manually calculated
        manually_calculated_result = 0.06604088

        assert form_3_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_t_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        t = -10  # -
        t_s = 2  # -
        h_0 = 200  # -

        with pytest.raises(ValueError):
            Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0)

    def test_raise_error_when_negative_t_s_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        t = 10  # -
        t_s = -2  # -
        h_0 = 200  # -

        with pytest.raises(ValueError):
            Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0)

    def test_raise_error_when_negative_h_0_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        t = 10  # -
        t_s = 2  # -
        h_0 = -200  # -

        with pytest.raises(ValueError):
            Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0)

    def test_raise_error_when_t_is_smaller_than_t_s(self) -> None:
        """Test a comparison."""
        # Example values
        t = 10  # -
        t_s = 12  # -
        h_0 = 200  # -

        with pytest.raises(ValueError):
            Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\beta_{ds}(t,t_s) = \frac{(t - t_s)}{(t - t_s) + 0.04 \sqrt{h_0^3}} = "
                r"\frac{(10.00 - 2.00)}{(10.00 - 2.00) + 0.04 \sqrt{200.00^3}} = 0.07",
            ),
            ("short", r"\beta_{ds}(t,t_s) = 0.07"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        t = 10  # -
        t_s = 2  # -
        h_0 = 200  # -

        # Object to test
        form_3_10_latex = Form3Dot10CoefficientAgeConcreteDryingShrinkage(t=t, t_s=t_s, h_0=h_0).latex()

        actual = {"complete": form_3_10_latex.complete, "short": form_3_10_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.74 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_74 import Form6Dot74MinimumCompressiveStressLevel
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot74MinimumCompressiveStressLevel:
    """Validation for formula 6.74 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        sigma_cd_min_equ = 10.0  # MPa
        f_cd_fat = 20.0  # MPa

        # Object to test
        formula = Form6Dot74MinimumCompressiveStressLevel(sigma_cd_min_equ=sigma_cd_min_equ, f_cd_fat=f_cd_fat)

        # Expected result, manually calculated
        manually_calculated_result = 0.5  # -

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("sigma_cd_min_equ", "f_cd_fat"),
        [
            (-10.0, 20.0),  # sigma_cd_min_equ is negative
            (10.0, -20.0),  # f_cd_fat is negative
            (10.0, 0.0),  # f_cd_fat is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, sigma_cd_min_equ: float, f_cd_fat: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot74MinimumCompressiveStressLevel(sigma_cd_min_equ=sigma_cd_min_equ, f_cd_fat=f_cd_fat)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"E_{cd,min,equ} = \frac{\sigma_{cd,min,equ}}{f_{cd,fat}} = \frac{10.000}{20.000} = 0.500 \ -",
            ),
            ("short", r"E_{cd,min,equ} = 0.500 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        sigma_cd_min_equ = 10.0  # MPa
        f_cd_fat = 20.0  # MPa

        # Object to test
        latex = Form6Dot74MinimumCompressiveStressLevel(sigma_cd_min_equ=sigma_cd_min_equ, f_cd_fat=f_cd_fat).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 3.11 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_11 import Form3Dot11AutogeneShrinkage


class TestForm3Dot11AutogeneShrinkage:
    """Validation for formula 3.11 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_as_t = 0.25  # -
        epsilon_ca_inf = 0.056  # -
        form_3_11 = Form3Dot11AutogeneShrinkage(beta_as_t=beta_as_t, epsilon_ca_inf=epsilon_ca_inf)

        # Expected result, manually calculated
        manually_calculated_result = 0.014

        assert form_3_11 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_beta_as_t_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        beta_as_t = -0.25  # -
        epsilon_ca_inf = 0.056  # -

        with pytest.raises(ValueError):
            Form3Dot11AutogeneShrinkage(beta_as_t=beta_as_t, epsilon_ca_inf=epsilon_ca_inf)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{ca}(t) = \beta_{as}(t) \cdot \epsilon_{ca}(\infty) = 0.250 \cdot 0.056 = 0.014",
            ),
            ("short", r"\epsilon_{ca}(t) = 0.014"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta_as_t = 0.25  # -
        epsilon_ca_inf = 0.056  # -

        # Object to test
        form_3_11_latex = Form3Dot11AutogeneShrinkage(beta_as_t=beta_as_t, epsilon_ca_inf=epsilon_ca_inf).latex()

        actual = {"complete": form_3_11_latex.complete, "short": form_3_11_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 6.5 from EN 1993-1-1:2005: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot5UnityCheckTensileStrength(Formula):
    """Class representing formula 6.5 for the unity check for tensile strength."""

    label = "6.5"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        n_ed: N,
        n_t_rd: N,
    ) -> None:
        r"""[$N_{Ed}/N_{t,Rd} \leq 1$] Unity check for tensile strength of an element in tension.

        EN 1993-1-1:2005 art.6.2.3(1) - Formula (6.5)

        Parameters
        ----------
        n_ed : N
            [$N_{Ed}$] Design value of the normal tensile force [$N$].
        n_t_rd : N
            [$N_{t,Rd}$] Design value of the resistance against tensile force [$N$].
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_t_rd = n_t_rd

    @staticmethod
    def _evaluate(
        n_ed: N,
        n_t_rd: N,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_t_rd=n_t_rd)
        raise_if_negative(n_ed=n_ed)
        return (n_ed / n_t_rd) <= 1

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.5."""
        _equation: str = r"\left( \frac{N_{Ed}}{N_{t,Rd}} \leq 1 \right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "N_{Ed}": f"{self.n_ed:.{n}f}",
                "N_{t,Rd}": f"{self.n_t_rd:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 7.8 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form7Dot8CrackWidth(Formula):
    r"""Class representing formula 7.8 for the calculation of the crack width [$w_k$]."""

    label = "7.8"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        s_r_max: MM,
        epsilon_sm_minus_epsilon_cm: DIMENSIONLESS,
    ) -> None:
        r"""[$w_k$] Calculation of the crack width [$mm$].

        EN 1992-1-1:2004 art.7.3.4(1) - Formula (7.8)

        Parameters
        ----------
        s_r_max : MM
            [$s_{r,max}$] Maximum crack spacing [$mm$].
        epsilon_sm_minus_epsilon_cm : DIMENSIONLESS
            [$\epsilon_{sm} - \epsilon_{cm}$] Difference between mean strain in reinforcement and mean strain in concrete.
            Followed from formula (7.9) [$-$].
        """
        super().__init__()
        self.s_r_max = s_r_max
        self.epsilon_sm_minus_epsilon_cm = epsilon_sm_minus_epsilon_cm

    @staticmethod
    def _evaluate(
        s_r_max: MM,
        epsilon_sm_minus_epsilon_cm: DIMENSIONLESS,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(s_r_max=s_r_max, epsilon_sm_minus_epsilon_cm=epsilon_sm_minus_epsilon_cm)

        return s_r_max * epsilon_sm_minus_epsilon_cm

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.8."""
        _equation: str = r"s_{r,max} \cdot (\epsilon_{sm} - \epsilon_{cm})"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"s_{r,max}": f"{self.s_r_max:.{n}f}",
                r"\epsilon_{sm} - \epsilon_{cm}": f"{self.epsilon_sm_minus_epsilon_cm:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"w_k",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm",
        )



"""Testing formula 6.70 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_70 import Form6Dot70FatigueDamageFactor
from blueprints.validations import LessOrEqualToZeroError, ListsNotSameLengthError, NegativeValueError


class TestForm6Dot70FatigueDamageFactor:
    """Validation for formula 6.70 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        n_delta_sigma_i = [100.0, 200.0, 300.0]
        capital_n_delta_sigma_i = [400.0, 500.0, 600.0]

        # Object to test
        formula = Form6Dot70FatigueDamageFactor(n_delta_sigma_i, capital_n_delta_sigma_i)

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("n_delta_sigma_i", "capital_n_delta_sigma_i"),
        [
            ([-100.0, 200.0, 300.0], [400.0, 500.0, 600.0]),  # n_delta_sigma_i contains negative value
            ([100.0, 200.0, 300.0], [0.0, 500.0, 600.0]),  # capital_n_delta_sigma_i contains zero
            ([100.0, 200.0, 300.0], [-400.0, 500.0, 600.0]),  # capital_n_delta_sigma_i contains negative value
            ([100.0, 200.0, 300.0], [400.0, 500.0]),  # n_delta_sigma_i and capital_n_delta_sigma_i have different lengths
        ],
    )
    def test_raise_error_when_negative_n_delta_sigma_i_is_given(self, n_delta_sigma_i: list, capital_n_delta_sigma_i: list) -> None:
        """Test invalid tests."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError, ListsNotSameLengthError)):
            Form6Dot70FatigueDamageFactor(n_delta_sigma_i, capital_n_delta_sigma_i)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \sum_{i} \frac{n(\Delta \sigma_i)}{N(\Delta \sigma_i)} < 1 \to \frac{100.000}{400.000} + "
                r"\frac{200.000}{500.000} + \frac{300.000}{600.000} < 1 \to \text{Not OK}",
            ),
            ("short", r"CHECK \to \text{Not OK}"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_delta_sigma_i = [100.0, 200.0, 300.0]
        capital_n_delta_sigma_i = [400.0, 500.0, 600.0]

        # Object to test
        latex = Form6Dot70FatigueDamageFactor(n_delta_sigma_i, capital_n_delta_sigma_i).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Triangular cross-section shape."""

from dataclasses import dataclass

from sectionproperties.pre import Geometry
from shapely import Polygon

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import MM


@dataclass(frozen=True)
class RightAngledTriangularCrossSection(CrossSection):
    """
    Class to represent a right-angled triangular with a right angle at the bottom left corner cross-section for geometric calculations.

    Parameters
    ----------
    base : MM
        The base length of the triangular cross-section.
    height : MM
        The height of the triangular cross-section.
    x : MM
        The x-coordinate of the 90-degree angle. Default is 0.
    y : MM
        The y-coordinate of the 90-degree angle. Default is 0.
    mirrored_horizontally : bool
        Whether the triangle is mirrored horizontally. Default is False.
    mirrored_vertically : bool
        Whether the triangle is mirrored vertically. Default is False.
    name : str
        The name of the rectangular cross-section, default is "Triangle".
    """

    base: MM
    height: MM
    x: MM = 0
    y: MM = 0
    mirrored_horizontally: bool = False
    mirrored_vertically: bool = False
    name: str = "Triangle"

    def __post_init__(self) -> None:
        """Post-initialization to validate the width and height."""
        if self.base < 0:
            raise ValueError(f"Base must be a positive value, but got {self.base}")
        if self.height < 0:
            raise ValueError(f"Height must be a positive value, but got {self.height}")

    @property
    def polygon(self) -> Polygon:
        """
        Shapely Polygon representing the right-angled triangular cross-section.

        Returns
        -------
        Polygon
            The shapely Polygon representing the triangle.
        """
        left_lower = (self.x, self.y)
        right_lower = (self.x + self.base, self.y)
        top = (self.x, self.y + self.height)

        if self.mirrored_horizontally:
            right_lower = (2 * left_lower[0] - right_lower[0], right_lower[1])
        if self.mirrored_vertically:
            top = (top[0], 2 * left_lower[1] - top[1])

        return Polygon([left_lower, right_lower, top])

    def geometry(
        self,
        mesh_size: MM | None = None,
    ) -> Geometry:
        """Return the geometry of the triangular cross-section.

        Parameters
        ----------
        mesh_size : MM
            Maximum mesh element area to be used within
            the Geometry-object finite-element mesh. If not provided, a default value will be used.

        Returns
        -------
        Geometry
            The Geometry object representing the triangular cross-section.
        """
        if mesh_size is None:
            minimum_mesh_size = 2.0
            mesh_length = max(min(self.base, self.height) / 20, minimum_mesh_size)
            mesh_size = mesh_length**2

        triangular = Geometry(geom=self.polygon)
        triangular.create_mesh(mesh_sizes=mesh_size)
        return triangular



"""Test the AZ enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.az import AZ


class TestAZ:
    """Tests for the AZ enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert AZ.AZ_18.value == (
            "AZ 18",
            630,
            380,
            9.5,
            9.5,
            348,
            55.4,
            150.4,
            74.4,
            34200,
            1800,
            1050,
            2104,
            118.1,
            15.07,
            1.35,
            "ArcelorMittal",
        )
        assert AZ.AZ_27_800.value == (
            "AZ 27-800",
            800,
            476,
            13.5,
            11.0,
            426,
            52.9,
            176.0,
            110.5,
            63570,
            2670,
            1550,
            3100,
            138.1,
            19.01,
            1.32,
            "ArcelorMittal",
        )

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "AZ 18" in [e.value[0] for e in AZ]
        assert "AZ 26" in [e.value[0] for e in AZ]
        assert "AZ 27-800" in [e.value[0] for e in AZ]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in AZ]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = AZ.AZ_18
        assert profile.alias == "AZ 18"
        assert profile.b_width_single_pile == 630
        assert profile.h_height_pile == 380
        assert profile.tf_flange_thickness == 9.5
        assert profile.tw_web_thickness == 9.5
        assert profile.bf_flange_width == 348
        assert profile.a_flange_angle == 55.4
        assert profile.a_cross_sectional_area == 150.4
        assert profile.gsp_mass_per_single_pile == 74.4
        assert profile.i_y_moment_inertia == 34200
        assert profile.w_el_y_elastic_section_modulus == 1800
        assert profile.s_y_static_moment == 1050
        assert profile.w_pl_y_plastic_section_modulus == 2104
        assert profile.gw_mass_per_m == 118.1
        assert profile.radius_of_gyration_y_y == 15.07
        assert profile.al_coating_area == 1.35
        assert profile.manufacturer == "ArcelorMittal"
        assert profile.sheet_pile_type == "Z-Section"

    def test_profile_count(self) -> None:
        """Test that all profiles are present."""
        # There should be 36 AZ profiles
        assert len(AZ) == 36

    def test_specific_profiles(self) -> None:
        """Test specific profiles from the JSON data."""
        # Test AZ 18-10/10
        profile = AZ.AZ_18_10_10
        assert profile.alias == "AZ 18-10/10"
        assert profile.h_height_pile == 381
        assert profile.tf_flange_thickness == 10.0
        assert profile.tw_web_thickness == 10.0

        # Test AZ 26
        profile = AZ.AZ_26
        assert profile.alias == "AZ 26"
        assert profile.h_height_pile == 427
        assert profile.tf_flange_thickness == 13.0
        assert profile.tw_web_thickness == 12.2

        # Test AZ 12-770
        profile = AZ.AZ_12_770
        assert profile.alias == "AZ 12-770"
        assert profile.b_width_single_pile == 770
        assert profile.h_height_pile == 344
        assert profile.a_flange_angle == 39.5



"""Formula 6.21 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot21ShearStressIOrHSection(Formula):
    r"""Class representing formula 6.21 for the calculation of [$\tau_{Ed}$]."""

    label = "6.21"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        v_ed: N,
        a_w: MM2,
        a_f: MM2,
    ) -> None:
        r"""[$\tau_{Ed}$] Calculation of the design elastic shear stress [$MPa$].
        For I- or H-sections the shear stress in the web may be taken with this equation.

        EN 1993-1-1:2005 art.6.2.6(5) - Formula (6.21)

        Parameters
        ----------
        v_ed : N
            [$V_{Ed}$] Design shear force [$N$].
        a_w : MM2
            [$A_w$] Area of the web [$mm^2$].
        a_f : MM2
            [$A_f$] Area of one flange [$mm^2$].
        """
        super().__init__()
        self.v_ed = v_ed
        self.a_w = a_w
        self.a_f = a_f

    @staticmethod
    def _evaluate(
        v_ed: N,
        a_w: MM2,
        a_f: MM2,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(v_ed=v_ed, a_f=a_f)
        raise_if_less_or_equal_to_zero(a_w=a_w)
        if not a_f / a_w >= 0.6:
            raise ValueError("A_f / A_w must be greater than or equal to 0.6")

        return v_ed / a_w

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.21."""
        _equation: str = r"\frac{V_{Ed}}{A_w} \text{ if } A_f / A_w \ge 0.6"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"V_{Ed}": f"{self.v_ed:.{n}f}",
                r"A_w": f"{self.a_w:.{n}f}",
                r"A_f": f"{self.a_f:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\tau_{Ed}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Testing sub-formula 3 for 8.8N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_8n import (
    SubForm8Dot8nFunctionX,
    SubForm8Dot8nFunctionY,
)
from blueprints.validations import NegativeValueError


class TestSubForm8Dot8nFunctionY:
    """Validation for sub-formula 8.8N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        x_function = 2

        sub_form_8_8n_3 = SubForm8Dot8nFunctionY(x_function=x_function)

        # Expected result, manually calculated
        manually_result = 0.112675

        assert sub_form_8_8n_3 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_x_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when x is negative."""
        # Example values
        x_function = -2

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nFunctionY(x_function=x_function)

    def test_integration_with_sub_form_8_8n_function_x(self) -> None:
        """Test the integration with sub-formula 8.8 for calculating function x."""
        # Example values
        cover = 60  # mm
        diameter_t = 16  # mm
        x_function = SubForm8Dot8nFunctionX(cover=cover, diameter_t=diameter_t)
        sub_form_8_8n_3 = SubForm8Dot8nFunctionY(x_function=x_function)

        # Expected result, manually calculated
        manually_result = 0.045314993

        assert sub_form_8_8n_3 == pytest.approx(expected=manually_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            ("complete", r"y = 0.015 + 0.14 \cdot e^{-0.18 \cdot x} = 0.015 + 0.14 \cdot e^{-0.18 \cdot 8.50} = 0.05"),
            ("short", r"y = 0.05"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        cover = 60  # mm
        diameter_t = 16  # mm
        x_function = SubForm8Dot8nFunctionX(cover=cover, diameter_t=diameter_t)

        # Object to test
        sub_form_8_8n_3_latex = SubForm8Dot8nFunctionY(x_function=x_function).latex()

        actual = {"complete": sub_form_8_8n_3_latex.complete, "short": sub_form_8_8n_3_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 5.40b from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot40bCheckLateralInstability(Formula):
    r"""Class representing formula 5.40b for checking lateral instability in transient situations."""

    label = "5.40b"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        l_0t: M,
        b: M,
        h: M,
    ) -> None:
        r"""Check the conditions to ignore second order effects in connection with lateral instability.

        EN 1992-1-1:2004 art.5.9(3) - Formula (5.40b)

        Parameters
        ----------
        l_0t : M
            [$l_{0t}$] is the distance between torsional restraints [$m$].
        b : M
            [$b$] is the width of compression flange [$m$].
        h : M
            [$h$] is the total depth of beam in central part of [$l_{0t}$] [$m$].
        """
        super().__init__()
        self.l_0t = l_0t
        self.b = b
        self.h = h

    @staticmethod
    def _evaluate(
        l_0t: M,
        b: M,
        h: M,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(b=b, h=h)
        raise_if_negative(l_0t=l_0t)

        return (l_0t / b <= 70 / (h / b) ** (1 / 3)) and (h / b <= 3.5)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.40b."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\left( \frac{l_{0t}}{b} \leq \frac{70}{\left( h/b \right)^{1/3}} \text{ and } \frac{h}{b} \leq 3.5 \right)",
            numeric_equation=rf"\left( \frac{{{self.l_0t:.{n}f}}}{{{self.b:.{n}f}}} \leq \frac{{70}}{{\left("
            rf" {self.h:.{n}f}/{self.b:.{n}f} \right)^{{1/3}}}} \text{{ and }} "
            rf"\frac{{{self.h:.{n}f}}}{{{self.b:.{n}f}}} \leq 3.5 \right)",
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 3.14 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA


class Form3Dot14StressStrainForShortTermLoading(Formula):
    """Class representing formula 3.14, which calculates the compressive stress-strength ratio."""

    source_document = EN_1992_1_1_2004
    label = "3.14"

    def __init__(
        self,
        k: DIMENSIONLESS,
        eta: DIMENSIONLESS,
    ) -> None:
        r"""[$\sigma_c / f_{cm}$] Compressive stress-strength ratio [$-$].

        EN 1992-1-1:2004 art.3.1.5(1) - Formula (3.14)

        Parameters
        ----------
        k : DIMENSIONLESS
            [$k$] [$-$].
            = 1.05 * Ecm * |$\epsilon_{c1}$| / fcm
            Use your own implementation of this formula or use the SubForm3Dot14K class.
        eta : DIMENSIONLESS
            [$\eta$] Strain - peak-strain ratio [$-$].
            = $\epsilon_c / \epsilon_{c1}$
            Use your own implementation of this formula or use the SubForm3Dot14Eta class.

        Returns
        -------
        None
        """
        super().__init__()
        self.k = k
        self.eta = eta

    @staticmethod
    def _evaluate(
        k: DIMENSIONLESS,
        eta: DIMENSIONLESS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if k < 0:
            raise ValueError(f"Invalid k: {k}. k cannot be negative")
        if eta < 0:
            raise ValueError(f"Invalid eta: {eta}. eta cannot be negative")
        return (k * eta - eta**2) / (1 + (k - 2) * eta)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.14."""
        return LatexFormula(
            return_symbol=r"\frac{\sigma_c}{f_{cm}}",
            result=f"{self:.{n}f}",
            equation=r"\frac{k \cdot \eta - \eta^2}{1 + (k-2) \cdot \eta}",
            numeric_equation=rf"\frac{{{self.k:.{n}f} \cdot {self.eta:.{n}f} - {self.eta:.{n}f}^2}}{{1 + ({self.k:.{n}f}-2) \cdot {self.eta:.{n}f}}}",
            comparison_operator_label="=",
        )


class SubForm3Dot14Eta(Formula):
    """Class representing sub-formula 1 for formula 3.14, which calculates eta."""

    source_document = EN_1992_1_1_2004
    label = "3.14"

    def __init__(
        self,
        epsilon_c: DIMENSIONLESS,
        epsilon_c1: DIMENSIONLESS,
    ) -> None:
        r"""[$\eta$] Strain - peak-strain ratio [$-$].

        EN 1992-1-1:2004 art.3.1.5(1) - η

        Parameters
        ----------
        epsilon_c : DIMENSIONLESS
            [$\epsilon_c$] Strain concrete [$-$].
        epsilon_c1 : DIMENSIONLESS
            [$\epsilon_{c1}$] Strain concrete at peak-stress following table 3.1 [$-$].
        """
        super().__init__()
        self.epsilon_c = epsilon_c
        self.epsilon_c1 = epsilon_c1

    @staticmethod
    def _evaluate(
        epsilon_c: DIMENSIONLESS,
        epsilon_c1: DIMENSIONLESS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        return epsilon_c / epsilon_c1

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.14 sub 1."""
        return LatexFormula(
            return_symbol=r"\eta",
            result=f"{self:.{n}f}",
            equation=r"\epsilon_c / \epsilon_{c1}",
            numeric_equation=rf"{self.epsilon_c:.{n}f} / {self.epsilon_c1:.{n}f}",
            comparison_operator_label="=",
        )


class SubForm3Dot14K(Formula):
    """Class representing sub-formula 2 for formula 3.14, which calculates k."""

    source_document = EN_1992_1_1_2004
    label = "3.14"

    def __init__(self, e_cm: MPA, epsilon_c1: DIMENSIONLESS, f_cm: MPA) -> None:
        r"""[$k$] [-].

        EN 1992-1-1:2004 art.3.1.5(1) - k

        Parameters
        ----------
        e_cm : MPA
            [$E_{cm}$] Elastic modulus concrete [$MPa$].
        epsilon_c1 : DIMENSIONLESS
            [$\epsilon_{c1}$] Strain concrete at peak-stress following table 3.1 [$-$].
        f_cm : MPA
            [$f_{cm}$] Compressive strength concrete [$MPa$].
        """
        super().__init__()
        self.e_cm = e_cm
        self.epsilon_c1 = epsilon_c1
        self.f_cm = f_cm

    @staticmethod
    def _evaluate(
        e_cm: MPA,
        epsilon_c1: DIMENSIONLESS,
        f_cm: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if e_cm < 0:
            raise ValueError(f"Invalid e_cm: {e_cm}. e_cm cannot be negative")
        if f_cm <= 0:
            raise ValueError(f"Invalid f_cm: {f_cm}. f_cm cannot be negative or zero")
        return 1.05 * e_cm * abs(epsilon_c1) / f_cm

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.14 sub 2."""
        return LatexFormula(
            return_symbol=r"k",
            result=f"{self:.{n}f}",
            equation=r"1.05 \cdot E_{cm} \cdot |\epsilon_{c1}| / f_{cm}",
            numeric_equation=rf"1.05 \cdot {self.e_cm:.{n}f} \cdot |{self.epsilon_c1:.{n}f}| / {self.f_cm:.{n}f}",
            comparison_operator_label="=",
        )



"""Sub-formula 3.28, 3.29 and 3.30 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA


class SubForm3Dot282930Mu(Formula):
    """Class representing sub-formula for 3.28, 3.29 and 3.30 for the calculation of μ."""

    label = "3.28 - 3.29 - 3.30"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        sigma_pi: MPA,
        f_pk: MPA,
    ) -> None:
        r"""[$\mu$] Ratio between initial pre-stress and characteristic tensile strength  [$\sigma_{pi} / f_{pk}$] [$-$].

        EN 1992-1-1:2004 art.3.3.2(7) - [$\mu$]

        Parameters
        ----------
        sigma_pi : MPA
            [$\sigma_{pi}$] Initial pre-stress [$MPa$]
        f_pk : MPA
            [$f_{pk}$] Characteristic tensile strength of pre-stress steel [$MPa$]

        Returns
        -------
        None
        """
        super().__init__()
        self.sigma_pi = sigma_pi
        self.f_pk = f_pk

    @staticmethod
    def _evaluate(
        sigma_pi: MPA,
        f_pk: MPA,
    ) -> DIMENSIONLESS:
        r"""Evaluates the formula, for more information see the __init__ method."""
        if f_pk < 0:
            raise ValueError(f"Invalid f_pk: {f_pk}. f_pk cannot be negative")
        return sigma_pi / f_pk

    def latex(self, n: int = 3) -> LatexFormula:
        r"""Returns LatexFormula object for [$\mu$] in formula 3.28 and 3.29 and 3.30."""
        return LatexFormula(
            return_symbol=r"\mu",
            result=f"{self:.{n}f}",
            equation=r"\sigma_{pi} / f_{pk}",
            numeric_equation=rf"{self.sigma_pi:.{n}f} / {self.f_pk:.{n}f}",
            comparison_operator_label="=",
        )



"""Formula A.1 from EN 1993-1-9:2005: Annex A - Determination of fatigue load parameters and verification formats."""

from blueprints.codes.eurocode.en_1993_1_9_2005 import EN_1993_1_9_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_lists_differ_in_length, raise_if_negative


class FormADot1DamageDuringDesignLife(Formula):
    """Class representing formula A.1 for the calculation of the damage during the design life [$D_d$]."""

    label = "A.1"
    source_document = EN_1993_1_9_2005

    def __init__(self, n_e: list[DIMENSIONLESS], n_r: list[DIMENSIONLESS]) -> None:
        r"""[$D_d$] The calculation of the damage during the design life [$-$].

        EN 1993-1-9:2005 art.A.5 - Formula (A.1)

        Parameters
        ----------
        n_e : list[DIMENSIONLESS]
            [$n_E$] Contains number of cycles associated with the stress range [$\gamma_{Ff} \cdot \Delta\sigma_i$]
            for each band i in the factored spectrum [$-$].
        n_r : list[DIMENSIONLESS]
            [$N_R$] Contains the endurance (in cycles) obtained from the factored [$\Delta\sigma_C / \gamma_{Mf} - n_r$] curve for
            each stress range of [$\gamma_{Ff} \cdot \Delta_i$] [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.n_e = n_e
        self.n_r = n_r

    @staticmethod
    def _evaluate(
        n_e: list[DIMENSIONLESS],
        n_r: list[DIMENSIONLESS],
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_lists_differ_in_length(n_e=n_e, n_r=n_r)
        raise_if_negative(n_e_min=min(n_e))
        raise_if_less_or_equal_to_zero(n_r_min=min(n_r))
        return sum([n_e[i] / n_r[i] for i in range(len(n_e))])

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula A.1."""
        return LatexFormula(
            return_symbol=r"D_d",
            result=f"{self:.{n}f}",
            equation=r"\sum_{i}^{n} \frac{n_{Ei}}{N_Ri}",
            numeric_equation="".join(rf"\frac{{{self.n_e[i]:.{n}f}}}{{{self.n_r[i]:.{n}f}}} + " for i in range(len(self.n_r)))[:-3],
            comparison_operator_label="=",
        )



"""Testing formula 6.36 from EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_36 import Form6Dot36MomentReduction
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot36MomentReduction:
    """Validation for formula 6.36 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        mpl_y_rd = 50000.0
        a = 0.4
        n = 0.5

        # Object to test
        formula = Form6Dot36MomentReduction(mpl_y_rd=mpl_y_rd, a=a, n=n)

        # Expected result, manually calculated
        manually_calculated_result = 31250.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation2(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        mpl_y_rd = 50000.0
        a = 0.4
        n = 0.0

        # Object to test
        formula = Form6Dot36MomentReduction(mpl_y_rd=mpl_y_rd, a=a, n=n)

        # Expected result, manually calculated
        manually_calculated_result = 50000.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("mpl_y_rd", "n"),
        [
            (-50000.0, 0.5),  # mpl_y_rd is negative
            (50000.0, -0.5),  # n is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, mpl_y_rd: float, n: float) -> None:
        """Test invalid negative values."""
        with pytest.raises(NegativeValueError):
            Form6Dot36MomentReduction(mpl_y_rd=mpl_y_rd, a=0.4, n=n)

    @pytest.mark.parametrize(
        "a",
        [
            2.0,  # denominator will be zero
            -1.0,  # a is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float) -> None:
        """Test invalid zero or negative values."""
        with pytest.raises((LessOrEqualToZeroError, NegativeValueError)):
            Form6Dot36MomentReduction(mpl_y_rd=10000.0, a=a, n=0.5)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{N,y,Rd} = \min\left(M_{pl,y,Rd}, M_{pl,y,Rd} \cdot (1 - n) / (1 - 0.5 \cdot a)\right) = "
                r"\min\left(50000.000, 50000.000 \cdot (1 - 0.500) / (1 - 0.5 \cdot 0.400)\right) = 31250.000 \ Nmm",
            ),
            (
                "complete_with_units",
                r"M_{N,y,Rd} = \min\left(M_{pl,y,Rd}, M_{pl,y,Rd} \cdot (1 - n) / (1 - 0.5 \cdot a)\right) = "
                r"\min\left(50000.000 \ Nmm, 50000.000 \ Nmm \cdot (1 - 0.500) / (1 - 0.5 \cdot 0.400)\right) = "
                r"31250.000 \ Nmm",
            ),
            ("short", r"M_{N,y,Rd} = 31250.000 \ Nmm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        mpl_y_rd = 50000.0
        a = 0.4
        n = 0.5

        # Object to test
        latex = Form6Dot36MomentReduction(mpl_y_rd=mpl_y_rd, a=a, n=n).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.8 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_8 import Form5Dot8EffectiveSpan
from blueprints.validations import NegativeValueError


class TestForm5Dot8EffectiveSpan:
    """Validation for formula 5.8 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        l_n = 5  # m
        a_1 = 0.5  # m
        a_2 = 0.35  # m

        # Object to test
        form_5_8 = Form5Dot8EffectiveSpan(l_n=l_n, a_1=a_1, a_2=a_2)

        # Expected result, manually calculated
        manually_calculated_result = 5.85  # kN

        assert form_5_8 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("l_n", "a_1", "a_2"),
        [
            (-5, 0.5, 0.35),
            (5, -0.5, 0.35),
            (5, 0.5, -0.35),
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, l_n: float, a_1: float, a_2: float) -> None:
        """Test negative values."""
        with pytest.raises(NegativeValueError):
            Form5Dot8EffectiveSpan(l_n=l_n, a_1=a_1, a_2=a_2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"l_{eff} = l_{n} + a_{1} + a_{2} = 5.000 + 0.500 + 0.350 = 5.850",
            ),
            ("short", r"l_{eff} = 5.850"),
            (
                "string",
                r"l_{eff} = l_{n} + a_{1} + a_{2} = 5.000 + 0.500 + 0.350 = 5.850",
            ),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        l_n = 5  # m
        a_1 = 0.5  # m
        a_2 = 0.35  # m

        # Object to test
        form_5_8_latex = Form5Dot8EffectiveSpan(
            l_n=l_n,
            a_1=a_1,
            a_2=a_2,
        ).latex()

        actual = {
            "complete": form_5_8_latex.complete,
            "short": form_5_8_latex.short,
            "string": str(form_5_8_latex),
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 4.2 of EN 1993-1-8:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_8_2005.chapter_4_welded_connections.formula_4_2 import Form4Dot2CheckWeldedConnection
from blueprints.validations import NegativeValueError


class TestForm4Dot2CheckWeldedConnection:
    """Validation for formula 4.2 from EN 1993-1-8:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        fw_ed = 50.0
        fw_rd = 100.0

        # Object to test
        formula = Form4Dot2CheckWeldedConnection(fw_ed=fw_ed, fw_rd=fw_rd)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("fw_ed", "fw_rd"),
        [
            (-10.0, 100.0),  # fw_ed is negative
            (50.0, -100.0),  # fw_rd is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, fw_ed: float, fw_rd: float) -> None:
        """Test invalid values where inputs are negative."""
        with pytest.raises(NegativeValueError):
            Form4Dot2CheckWeldedConnection(fw_ed=fw_ed, fw_rd=fw_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to F_{w,Ed} \leq F_{w,Rd} \to 50.000 \leq 100.000 \to OK",
            ),
            (
                "complete_with_units",
                r"CHECK \to F_{w,Ed} \leq F_{w,Rd} \to 50.000 \ N \leq 100.000 \ N \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        fw_ed = 50.0
        fw_rd = 100.0

        # Object to test
        latex = Form4Dot2CheckWeldedConnection(fw_ed=fw_ed, fw_rd=fw_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
            "complete_with_units": latex.complete_with_units,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 9.9 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_9 import Form9Dot9MaximumSpacingSeriesOfLinks
from blueprints.validations import NegativeValueError


class TestForm9Dot9MaximumSpacingSeriesOfLinks:
    """Validation for formula 9.9 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        d = 100  # mm
        alpha = 85  # deg
        form_9_9 = Form9Dot9MaximumSpacingSeriesOfLinks(d=d, alpha=alpha)

        # Expected result, manually calculated
        manually_calculated_result = 81.56164976

        assert form_9_9 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_d_is_given(self) -> None:
        """Test if error is given when d is negative."""
        d = -100  # mm
        alpha = 85  # deg

        with pytest.raises(NegativeValueError):
            Form9Dot9MaximumSpacingSeriesOfLinks(d=d, alpha=alpha)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"s_{max} = 0.75 \cdot d \cdot \left( 1 + cot(\alpha) \right) = 0.75 \cdot 100.00 \cdot \left( 1 + cot(85.00) \right) = 81.56",
            ),
            ("short", r"s_{max} = 81.56"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        d = 100  # mm
        alpha = 85  # deg

        # Object to test
        form_9_9n_latex = Form9Dot9MaximumSpacingSeriesOfLinks(d=d, alpha=alpha).latex()

        actual = {"complete": form_9_9n_latex.complete, "short": form_9_9n_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.73 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_73 import Form6Dot73StressRatio
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot73StressRatio:
    """Validation for formula 6.73 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        e_cd_min_equ = 10.0
        e_cd_max_equ = 20.0

        # Object to test
        formula = Form6Dot73StressRatio(e_cd_min_equ=e_cd_min_equ, e_cd_max_equ=e_cd_max_equ)

        # Expected result, manually calculated
        manually_calculated_result = 0.5

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("e_cd_min_equ", "e_cd_max_equ"),
        [
            (-10.0, 20.0),  # e_cd_min_equ is negative
            (10.0, -20.0),  # e_cd_max_equ is negative
            (10.0, 0.0),  # e_cd_max_equ is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, e_cd_min_equ: float, e_cd_max_equ: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot73StressRatio(e_cd_min_equ=e_cd_min_equ, e_cd_max_equ=e_cd_max_equ)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"R_{equ} = \frac{E_{cd,min,equ}}{E_{cd,max,equ}} = \frac{10.000}{20.000} = 0.500",
            ),
            ("short", r"R_{equ} = 0.500"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_cd_min_equ = 10.0
        e_cd_max_equ = 20.0

        # Object to test
        latex = Form6Dot73StressRatio(e_cd_min_equ=e_cd_min_equ, e_cd_max_equ=e_cd_max_equ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 3.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_1 import Form6Dot1DesignShearStrength


class TestForm6Dot1DesignShearStrength:
    """Validation for formula 6.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        v_rd_s = 10  # kN
        v_ccd = 10  # kN
        v_td = 10  # kN
        form = Form6Dot1DesignShearStrength(v_rd_s=v_rd_s, v_ccd=v_ccd, v_td=v_td)

        # Expected result, manually calculated
        expected = 30

        assert form == expected

    def test_raise_error_when_negative_v_rd_s_is_given(self) -> None:
        """Test a negative value for v_rd_s."""
        # Example values
        v_rd_s = -10
        v_ccd = 10
        v_td = 10

        with pytest.raises(ValueError):
            Form6Dot1DesignShearStrength(v_rd_s=v_rd_s, v_ccd=v_ccd, v_td=v_td)

    def test_raise_error_when_negative_v_ccd_is_given(self) -> None:
        """Test a negative value for v_ccd."""
        # Example values
        v_rd_s = 10
        v_ccd = -10
        v_td = 10

        with pytest.raises(ValueError):
            Form6Dot1DesignShearStrength(v_rd_s=v_rd_s, v_ccd=v_ccd, v_td=v_td)

    def test_raise_error_when_negative_v_td_is_given(self) -> None:
        """Test a negative value for v_td."""
        # Example values
        v_rd_s = 10
        v_ccd = 10
        v_td = -10

        with pytest.raises(ValueError):
            Form6Dot1DesignShearStrength(v_rd_s=v_rd_s, v_ccd=v_ccd, v_td=v_td)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            ("complete", r"V_{Rd} = V_{Rd,s} + V_{ccd} + V_{td} = 2.000 + 3.000 + 4.000 = 9.000"),
            ("short", r"V_{Rd} = 9.000"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_rd_s = 2.0
        v_ccd = 3.0
        v_td = 4.0

        # Object to test
        form_6_1_latex = Form6Dot1DesignShearStrength(v_rd_s=v_rd_s, v_ccd=v_ccd, v_td=v_td).latex()

        actual = {"complete": form_6_1_latex.complete, "short": form_6_1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 6.36 from EN 1993-1-1:2005: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, NMM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot36MomentReduction(Formula):
    r"""Class representing formula 6.36 for the calculation of reduced bending moment when axially loaded."""

    label = "6.36"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        mpl_y_rd: NMM,
        a: DIMENSIONLESS,
        n: DIMENSIONLESS,
    ) -> None:
        r"""[$M_{N,y,Rd}$] Reduced bending moment [$Nmm$].

        EN 1993-1-1:2005 art.6.2.9.1(5) - Formula (6.36)

        Parameters
        ----------
        mpl_y_rd : NMM
            [$M_{pl,y,Rd}$] Plastic bending moment about the y-axis [$Nmm$].
        a : DIMENSIONLESS
            Reduction factor for cross-sectional area, see formula 6.38a (`Form6Dot38A`) [-].
        n : DIMENSIONLESS
            Axial force ratio, see formula 6.38n (`Form6Dot38N`) [-].
        """
        super().__init__()
        self.mpl_y_rd = mpl_y_rd
        self.a = a
        self.n = n

    @staticmethod
    def _evaluate(
        mpl_y_rd: NMM,
        a: DIMENSIONLESS,
        n: DIMENSIONLESS,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(mpl_y_rd=mpl_y_rd, n=n, a=a)

        denominator = 1 - 0.5 * a
        raise_if_less_or_equal_to_zero(denominator=denominator)

        result = mpl_y_rd * (1 - n) / denominator
        return min(result, mpl_y_rd)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.36."""
        _equation: str = r"\min\left(M_{pl,y,Rd}, M_{pl,y,Rd} \cdot (1 - n) / (1 - 0.5 \cdot a)\right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "M_{pl,y,Rd}": f"{self.mpl_y_rd:.{n}f}",
                " n": f" {self.n:.{n}f}",
                "a": f"{self.a:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                "M_{pl,y,Rd}": rf"{self.mpl_y_rd:.{n}f} \ Nmm",
                " n": rf" {self.n:.{n}f}",
                "a": rf"{self.a:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"M_{N,y,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Testing formula 7.16.a and 7.16.b of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_16_ab import (
    Form7Dot16abSpanDepthRatio,
    Form7Dot16ReferenceReinforcementRatio,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot16abSpanDepthRatio:
    """Validation for formula 7.16a/b from EN 1992-1-1:2004."""

    def test_evaluation_rho_less_than_rho_0(self) -> None:
        """Tests the evaluation of the result when rho <= rho_0."""
        # Example values
        capital_k = 1.0
        f_ck = 30.0
        rho = 0.001
        rho_0 = 0.002
        rho_prime = 0.0005

        # Object to test
        formula = Form7Dot16abSpanDepthRatio(capital_k=capital_k, f_ck=f_ck, rho=rho, rho_0=rho_0, rho_prime=rho_prime)

        # Expected result, manually calculated
        manually_calculated_result = 44.9587985653203

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_rho_greater_than_rho_0(self) -> None:
        """Tests the evaluation of the result when rho > rho_0."""
        # Example values
        capital_k = 1.0
        f_ck = 30.0
        rho = 0.003
        rho_0 = 0.002
        rho_prime = 0.0005

        # Object to test
        formula = Form7Dot16abSpanDepthRatio(capital_k=capital_k, f_ck=f_ck, rho=rho, rho_0=rho_0, rho_prime=rho_prime)

        # Expected result, manually calculated
        manually_calculated_result = 17.80088842235581

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("capital_k", "f_ck", "rho", "rho_0", "rho_prime"),
        [
            (-1.0, 30.0, 0.001, 0.002, 0.0),  # capital_k is negative
            (1.0, -30.0, 0.001, 0.002, 0.0),  # f_ck is negative
            (1.0, 30.0, -0.001, 0.002, 0.0),  # rho is negative
            (1.0, 30.0, 0.001, -0.002, 0.0),  # rho_0 is negative
            (1.0, 30.0, 0.001, 0.002, -0.001),  # rho_prime is negative
            (1.0, 30.0, 0.0, 0.002, 0.0),  # rho is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, capital_k: float, f_ck: float, rho: float, rho_0: float, rho_prime: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot16abSpanDepthRatio(capital_k=capital_k, f_ck=f_ck, rho=rho, rho_0=rho_0, rho_prime=rho_prime)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\frac{l}{d} = \begin{cases} K \cdot \left[11 + 1.5 \cdot \sqrt{f_{ck}} \cdot "
                r"\frac{\rho_0}{\rho} + 3.2 \cdot \sqrt{f_{ck}} \cdot \left(\frac{\rho_0}{\rho} - 1\right)^{3/2}\right] & "
                r"\text{if } \rho \leq \rho_0 \\ K \cdot \left[11 + 1.5 \cdot \sqrt{f_{ck}} \cdot "
                r"\frac{\rho_0}{\rho - \rho'} + \frac{1}{12} \cdot \sqrt{f_{ck}} \cdot \sqrt{\frac{\rho'}"
                r"{\rho_0}}\right] & \text{if } \rho > \rho_0 \end{cases} = "
                r"\begin{cases} 1.000 \cdot \left[11 + 1.5 \cdot \sqrt{30.000} \cdot \frac{0.0020}{0.0010} + "
                r"3.2 \cdot \sqrt{30.000} \cdot \left(\frac{0.0020}{0.0010} - 1\right)^{3/2}\right] & "
                r"\text{if } 0.0010 \leq 0.0020 \\ 1.000 \cdot \left[11 + 1.5 \cdot \sqrt{30.000} "
                r"\cdot \frac{0.0020}{0.0010 - 0.0005} + \frac{1}{12} \cdot \sqrt{30.000} \cdot "
                r"\sqrt{\frac{0.0005}{0.0020}}\right] & \text{if } 0.0010 > 0.0020 \end{cases} = 44.959 \ -",
            ),
            ("short", r"\frac{l}{d} = 44.959 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        capital_k = 1.0
        f_ck = 30.0
        rho = 0.0010
        rho_0 = 0.0020
        rho_prime = 0.0005

        # Object to test
        latex = Form7Dot16abSpanDepthRatio(capital_k=capital_k, f_ck=f_ck, rho=rho, rho_0=rho_0, rho_prime=rho_prime).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm7Dot16ReferenceReinforcementRatio:
    """Validation for reference reinforcement ratio calculation from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        f_ck = 30.0

        formula = Form7Dot16ReferenceReinforcementRatio(f_ck=f_ck)
        manually_calculated_result = 0.00547722557  # Example manually calculated result

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        "f_ck",
        [
            -30.0,  # f_ck is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, f_ck: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form7Dot16ReferenceReinforcementRatio(f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\rho_0 = \sqrt{f_{ck}} \cdot 10^{-3}"
                r" = \sqrt{30.000} \cdot 10^{-3} = 0.005477 \ -",
            ),
            ("short", r"\rho_0 = 0.005477 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        f_ck = 30.0

        latex = Form7Dot16ReferenceReinforcementRatio(f_ck=f_ck).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.8 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_8 import Form6Dot8ShearResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot8ShearResistance:
    """Validation for formula 6.8 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_sw = 100.0
        s = 200.0
        z = 300.0
        f_ywd = 400.0
        theta = 45.0

        # Object to test
        formula = Form6Dot8ShearResistance(a_sw=a_sw, s=s, z=z, f_ywd=f_ywd, theta=theta)

        # Expected result, manually calculated
        manually_calculated_result = 60000.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_sw", "s", "z", "f_ywd", "theta"),
        [
            (-100.0, 200.0, 300.0, 400.0, 45.0),  # a_sw is negative
            (100.0, -200.0, 300.0, 400.0, 45.0),  # s is negative
            (100.0, 200.0, -300.0, 400.0, 45.0),  # z is negative
            (100.0, 200.0, 300.0, -400.0, 45.0),  # f_ywd is negative
            (100.0, 200.0, 300.0, 400.0, -45.0),  # theta is negative
            (100.0, 0.0, 300.0, 400.0, 45.0),  # s is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_sw: float, s: float, z: float, f_ywd: float, theta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot8ShearResistance(a_sw=a_sw, s=s, z=z, f_ywd=f_ywd, theta=theta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Rd,s} = \frac{A_{sw}}{s} \cdot z \cdot f_{ywd} \cdot \cot(\theta) = "
                r"\frac{100.000}{200.000} \cdot 300.000 \cdot 400.000 \cdot \cot(45.000) = 60000.000 \ N",
            ),
            ("short", r"V_{Rd,s} = 60000.000 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_sw = 100.0
        s = 200.0
        z = 300.0
        f_ywd = 400.0
        theta = 45.0

        # Object to test
        latex = Form6Dot8ShearResistance(a_sw=a_sw, s=s, z=z, f_ywd=f_ywd, theta=theta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.43 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_negative


class Form5Dot43InitialPrestressForce(Formula):
    r"""Class representing formula 5.43 for the calculation of the initial prestress force, [$P_{m0}(x)$]."""

    label = "5.43"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_p: MM2,
        k_7: DIMENSIONLESS,
        f_pk: MPA,
        k_8: DIMENSIONLESS,
        f_p0_1k: MPA,
    ) -> None:
        r"""[$P_{m0}(x)$] Initial prestress force [$N$].

        EN 1992-1-1:2004 art.5.10.3(2) - Formula (5.43)

        Parameters
        ----------
        a_p : MM2
            [$A_{p}$] Area of prestressing steel [$mm^2$].
        k_7 : DIMENSIONLESS
            [$k_{7}$] Coefficient for characteristic tensile strength, recommended value is 0.75 [$-$].
        f_pk : MPA
            [$f_{pk}$] Characteristic tensile strength of prestressing steel [$MPa$].
        k_8 : DIMENSIONLESS
            [$k_{8}$] Coefficient for 0.1% proof stress, recommended value is 0.85 [$-$].
        f_p0_1k : MPA
            [$f_{p0,1k}$] 0.1% proof stress of prestressing steel [$MPa$].
        """
        super().__init__()
        self.a_p = a_p
        self.k_7 = k_7
        self.f_pk = f_pk
        self.k_8 = k_8
        self.f_p0_1k = f_p0_1k

    @staticmethod
    def _evaluate(
        a_p: MM2,
        k_7: DIMENSIONLESS,
        f_pk: MPA,
        k_8: DIMENSIONLESS,
        f_p0_1k: MPA,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            a_p=a_p,
            k_7=k_7,
            f_pk=f_pk,
            k_8=k_8,
            f_p0_1k=f_p0_1k,
        )

        return a_p * min(k_7 * f_pk, k_8 * f_p0_1k)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.43."""
        return LatexFormula(
            return_symbol=r"P_{m0}(x)",
            result=f"{self:.{n}f}",
            equation=r"A_{p} \cdot \min \left(k_7 \cdot f_{pk} ; k_8 \cdot f_{p0.1k} \right)",
            numeric_equation=rf"{self.a_p:.{n}f} \cdot \min \left({self.k_7:.{n}f} \cdot {self.f_pk:.{n}f} ;"
            rf" {self.k_8:.{n}f} \cdot {self.f_p0_1k:.{n}f} \right)",
            comparison_operator_label="=",
            unit="N",
        )



"""Formula 6.18 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot18DesignPlasticShearResistance(Formula):
    r"""Class representing formula 6.18 for the calculation of [$V_{pl,Rd}$]."""

    label = "6.18"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a_v: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> None:
        r"""[$V_{pl,Rd}$] Calculation of the design plastic shear resistance [$N$].

        EN 1993-1-1:2005 art.6.2.6(2) - Formula (6.18)

        Parameters
        ----------
        a_v : MM2
            [$A_v$] Shear area, to be taken from a subformula from 6.18 [$mm^2$].
        f_y : MPA
            [$f_y$] Yield strength of the material [$MPa$].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial safety factor for resistance of cross-sections.
        """
        super().__init__()
        self.a_v = a_v
        self.f_y = f_y
        self.gamma_m0 = gamma_m0

    @staticmethod
    def _evaluate(
        a_v: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a_v=a_v, f_y=f_y)
        raise_if_less_or_equal_to_zero(gamma_m0=gamma_m0)

        return (a_v * (f_y / np.sqrt(3))) / gamma_m0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18."""
        _equation: str = r"\frac{A_v \cdot (f_y / \sqrt{3})}{\gamma_{M0}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A_v": f"{self.a_v:.{n}f}",
                r"f_y": f"{self.f_y:.{n}f}",
                r"\gamma_{M0}": f"{self.gamma_m0:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"V_{pl,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="N",
        )



"""Testing table 4.2 of NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

import pytest

from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.table_4_2 import Table4Dot2MinimumCoverWithRegardToBond
from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_14 import (
    Form8Dot14EquivalentDiameterBundledBars,
)
from blueprints.type_alias import MM
from blueprints.validations import LessOrEqualToZeroError


class TestTable4Dot2MinimumCoverWithRegardToBond:
    """Validation for table 4.2 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

    @pytest.mark.parametrize(
        ("diameter", "nominal_max_aggregate_size_greater_than_32_mm", "expected_result"),
        [
            (25, True, 30),
            (25, False, 25),
        ],
    )
    def test_evaluation(self, diameter: MM, nominal_max_aggregate_size_greater_than_32_mm: bool, expected_result: MM) -> None:
        """Test the evaluation of the result."""
        form_4_2 = Table4Dot2MinimumCoverWithRegardToBond(
            diameter=diameter,
            nominal_max_aggregate_size_greater_than_32_mm=nominal_max_aggregate_size_greater_than_32_mm,
        )

        assert form_4_2 == pytest.approx(expected=expected_result, rel=1e-4)

    def test_evaluation_using_form_8_14(self) -> None:
        """Test the evaluation of the result using formula 8.14 (Form8Dot14EquivalentDiameterBundledBars)."""
        form_4_2 = Table4Dot2MinimumCoverWithRegardToBond(
            diameter=Form8Dot14EquivalentDiameterBundledBars(diameter=20, n_b=4),
            nominal_max_aggregate_size_greater_than_32_mm=True,
        )

        expected_result = 45

        assert form_4_2 == pytest.approx(expected=expected_result, rel=1e-4)

    def test_raise_error_when_negative_diameter_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        diameter = -7
        nominal_max_aggregate_size_greater_than_32_mm = False

        with pytest.raises(LessOrEqualToZeroError):
            Table4Dot2MinimumCoverWithRegardToBond(
                diameter=diameter,
                nominal_max_aggregate_size_greater_than_32_mm=nominal_max_aggregate_size_greater_than_32_mm,
            )

    def test_raise_error_when_invalid_boolean_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        diameter = 20
        nominal_max_aggregate_size_greater_than_32_mm = 1

        with pytest.raises(TypeError) as exc_info:
            Table4Dot2MinimumCoverWithRegardToBond(
                diameter=diameter,
                nominal_max_aggregate_size_greater_than_32_mm=nominal_max_aggregate_size_greater_than_32_mm,  # type: ignore[arg-type]
            )

        assert str(exc_info.value) == "The parameter 'nominal_max_aggregate_size_greater_than_32_mm' must be a boolean."

    @pytest.mark.parametrize(
        ("nominal_max_aggregate_size_greater_than_32_mm", "representation", "expected_result"),
        [
            (True, "complete", r"c_{min,b} = (equivalent) rebar diameter + 5 = 20 + 5 = 25"),
            (True, "short", "c_{min,b} = 25"),
            (False, "complete", r"c_{min,b} = (equivalent) rebar diameter = 20 = 20"),
            (False, "short", "c_{min,b} = 20"),
        ],
    )
    def test_latex(self, nominal_max_aggregate_size_greater_than_32_mm: bool, representation: str, expected_result: str) -> None:
        """Test the latex implementation."""
        diameter = 20

        form = Table4Dot2MinimumCoverWithRegardToBond(
            diameter=diameter, nominal_max_aggregate_size_greater_than_32_mm=nominal_max_aggregate_size_greater_than_32_mm
        ).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



r"""Formula 5.38b from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot38bCheckRelativeEccentricityRatio(Formula):
    r"""Class representing formula 5.38b for check of relative eccentricity ratio."""

    label = "5.38b"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        e_y: MM,
        e_z: MM,
        b_eq: MM,
        h_eq: MM,
    ) -> None:
        r"""Check the excentricity of the loads in x and y direction.

        EN 1992-1-1:2004 art.5.8.9(3) - Formula (5.38b)

        Parameters
        ----------
        e_y : MM
            [$e_{y}$] Eccentricity along y-axis [$mm$].
        e_z : MM
            [$e_{z}$] Eccentricity along z-axis [$mm$].
        b_eq : MM
            [$b_{eq}$] Equivalent width [$mm$].
        h_eq : MM
            [$h_{eq}$] Equivalent depth [$mm$].
        """
        super().__init__()
        self.e_y = e_y
        self.e_z = e_z
        self.b_eq = b_eq
        self.h_eq = h_eq

    @staticmethod
    def _evaluate(
        e_y: MM,
        e_z: MM,
        b_eq: MM,
        h_eq: MM,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(e_y=e_y, e_z=e_z)
        raise_if_less_or_equal_to_zero(b_eq=b_eq, h_eq=h_eq)

        # The formula is simplified to avoid division by zero.
        if e_y == 0 or e_z == 0:
            return True

        return (e_y / h_eq) / (e_z / b_eq) <= 0.2 or (e_z / b_eq) / (e_y / h_eq) <= 0.2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.38b."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\left(\frac{e_{y}/h_{eq}}{e_{z}/b_{eq}} \leq 0.2 \text{ or } \frac{e_{z}/b_{eq}}{e_{y}/h_{eq}} \leq 0.2 \right)",
            numeric_equation=rf"\left(\frac{{{self.e_y:.{n}f}/{self.h_eq:.{n}f}}}{{{self.e_z:.{n}f}/{self.b_eq:.{n}f}}} "
            rf"\leq 0.2 \text{{ or }} \frac{{{self.e_z:.{n}f}/{self.b_eq:.{n}f}}}{{{self.e_y:.{n}f}/{self.h_eq:.{n}f}}} \leq 0.2 \right)",
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 5.25 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form5Dot25AxialForceCorrectionFactor(Formula):
    """Class representing formula 5.25 for the calculation of the axial force correction factor [$k_{2}$]."""

    label = "5.25"
    source_document = EN_1992_1_1_2004

    def __init__(self, n: DIMENSIONLESS) -> None:
        r"""[$k_{2}$] Axial force correction factor.

        EN 1992-1-1:2004 art.5.8.6(3) - Formula (5.25)

        Parameters
        ----------
        n : DIMENSIONLESS
            [$n$] Relative axial force, [$N_{ed} / (A_{c} * f_{cd})$] [-].
        """
        super().__init__()
        self.n = n

    @staticmethod
    def _evaluate(n: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n=n)
        k2 = n * 0.30
        return min(k2, 0.20)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.25."""
        return LatexFormula(
            return_symbol=r"k_{2}",
            result=f"{self:.{n}f}",
            equation=r"\min(0.20; n \cdot 0.30)",
            numeric_equation=rf"\min(0.20; {self.n:.{n}f} \cdot 0.30)",
            comparison_operator_label="=",
        )



"""Formula 6.46 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot46BetaCorner(Formula):
    r"""Class representing formula 6.46 for the calculation of [$\beta$] for corner column connections."""

    label = "6.46"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        u1: MM,
        u1_star: MM,
    ) -> None:
        r"""[$\beta$] Calculation of [$\beta$].

        EN 1992-1-1:2004 art.6.4.3(3) - Formula (6.46)

        Parameters
        ----------
        u1 : MM
            [$u_1$] Basic control perimeter [$mm$].
        u1_star : MM
            [$u_{1^*}$] Reduced basic control perimeter [$mm$].
        """
        super().__init__()
        self.u1 = u1
        self.u1_star = u1_star

    @staticmethod
    def _evaluate(
        u1: MM,
        u1_star: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(u1=u1)
        raise_if_less_or_equal_to_zero(u1_star=u1_star)

        return u1 / u1_star

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.46."""
        _equation: str = r"\frac{u_1}{u_{1^*}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"u_1": f"{self.u1:.{n}f}",
                r"u_{1^*}": f"{self.u1_star:.{n}f}",
            },
            True,
        )
        return LatexFormula(
            return_symbol=r"\beta",
            result=f"{self._evaluate(self.u1, self.u1_star):.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Definitions and base classes for constants for the calculation of nominal concrete cover according to art. 4.4.1 from EN 1992-1-1:2004."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

from blueprints.type_alias import MM


class AbrasionClass(Enum):
    """Enum representing the abrasion class of the concrete surface.

    According to art. 4.4.1.2 (13) from EN 1992-1-1:2004
    """

    NA = "Not applicable"
    XM1 = "XM1"
    XM2 = "XM2"
    XM3 = "XM3"


class CastingSurface(Enum):
    """Enum representing the casting surface of the concrete.

    According to art. 4.4.1.3 (4) from EN 1992-1-1:2004
    """

    PERMANENTLY_EXPOSED = "Permanently exposed"
    FORMWORK = "Formwork"
    PREPARED_GROUND = "Prepared ground (including blinding)"
    DIRECTLY_AGAINST_SOIL = "Directly against soil"


@dataclass(frozen=True)
class NominalConcreteCoverConstantsBase(ABC):
    """Base class for constants for the calculation of nominal concrete cover."""

    CODE_PREFIX: str  # Prefix for the code representing the country, e.g. "NEN-, DIN-, BS-, etc."
    CODE_SUFFIX: str  # Suffix for the code representing the publication year, e.g. "+C2:2011"
    COVER_INCREASE_FOR_UNEVEN_SURFACE: MM
    COVER_INCREASE_FOR_ABRASION_CLASS: dict[AbrasionClass, MM]
    DEFAULT_DELTA_C_DEV: MM

    @staticmethod
    @abstractmethod
    def minimum_cover_with_regard_to_casting_surface(c_min_dur: MM, casting_surface: CastingSurface) -> MM:
        """Calculate the minimum cover with regard to casting surface according to art. 4.4.1.3 (4) from EN 1992-1-1."""

    @staticmethod
    @abstractmethod
    def minimum_cover_with_regard_to_casting_surface_latex(casting_surface: CastingSurface) -> str:
        """LateX representation of minimum cover with regard to casting surface according to art. 4.4.1.3 (4) from EN 1992-1-1."""



"""Formula 3.16 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA


class Form3Dot16DesignValueTensileStrength(Formula):
    """Class representing formula 3.16 for the calculation of the concrete tensile strength design value."""

    label = "3.16"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_ct: DIMENSIONLESS,
        f_ctk_0_05: MPA,
        gamma_c: DIMENSIONLESS,
    ) -> None:
        r"""[$f_{cd}$] Design value concrete tensile strength [$MPa$].

        EN 1992-1-1:2004 art.3.1.6(2) - Formula (3.16)

        Parameters
        ----------
        alpha_ct : DIMENSIONLESS
            [$\alpha_{ct}$] Coefficient taking long term effects on tensile strength into
            account and unfavorable effect due to positioning loading [$-$]
            See national appendix. Recommended value: 1.0
        f_ctk_0_05 : MPA
            [$f_{ctk,0.05}$] Characteristic tensile strength 5% [$MPa$].
        gamma_c : DIMENSIONLESS
            [$\gamma_{c}$] Partial safety factor concrete, see 2.4.2.4 [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.alpha_ct = alpha_ct
        self.f_ctk_0_05 = f_ctk_0_05
        self.gamma_c = gamma_c

    @staticmethod
    def _evaluate(
        alpha_ct: DIMENSIONLESS,
        f_ctk_0_05: MPA,
        gamma_c: DIMENSIONLESS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if alpha_ct < 0:
            raise ValueError(f"Invalid alpha_ct: {alpha_ct}. alpha_ct cannot be negative")
        if f_ctk_0_05 < 0:
            raise ValueError(f"Invalid f_ctk_0_05: {f_ctk_0_05}. f_ctk_0_05 cannot be negative")
        if gamma_c <= 0:
            raise ValueError(f"Invalid gamma_c: {gamma_c}. gamma_c cannot be negative or zero")
        return alpha_ct * f_ctk_0_05 / gamma_c

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.16."""
        return LatexFormula(
            return_symbol=r"f_{ctd}",
            result=f"{self:.{n}f}",
            equation=r"\alpha_{ct} \cdot f_{ctk,0.05} / \gamma_C",
            numeric_equation=rf"{self.alpha_ct:.{n}f} \cdot {self.f_ctk_0_05:.{n}f} / {self.gamma_c:.{n}f}",
            comparison_operator_label="=",
        )



"""Formula 5.18 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, MM4, MPA, M
from blueprints.unit_conversion import M_TO_MM, N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot18ComparisonGeneralSecondOrderEffects:
    """Class representing comparison 5.18 for general second-order effects in buildings."""

    label = "5.18"
    source_document = EN_1992_1_1_2004

    def __init__(self, f_ved: KN, k_1: DIMENSIONLESS, n_s: DIMENSIONLESS, length: M, e_cd: MPA, i_c: MM4) -> None:
        r"""[$CHECK$] Criteria met, based on damage accumulation.

        EN 1992-1-1:2004 art.5.8.3.3(1) - Formula (5.18)

        Parameters
        ----------
        f_ved : KN
            [$F_{v,ed}$] Total vertical load (on braced and bracing members) [$kN$].
        k_1 : DIMENSIONLESS
            [$k_1$] The value of k1 for use in a Country may be found in its National Annex. Recommend value is
            0.31 [$-$].
        n_s : DIMENSIONLESS
            [$n_s$] is the total number of storeys [$-$].
        length : M
            [$L$] is the total height of the building above level of moment restraint. [$m$].
        e_cd : MPa
            [$E_{cd}$] is the design value of the modulus of elasticity of concrete. [$MPa$].
        i_c : MM4
            [$I_c$] is the second moment of area (uncracked concrete section) of bracing member(s). [$mm^4$].
        """
        self.f_ved = f_ved
        self.k_1 = k_1
        self.n_s = n_s
        self.length = length
        self.e_cd = e_cd
        self.i_c = i_c

    @property
    def left_hand_side(self) -> KN:
        """Calculate the left hand side of the comparison.

        Returns
        -------
        KN
            Left hand side of the comparison.
        """
        return self.f_ved

    @property
    def right_hand_side(self) -> KN:
        """Calculate the right hand side of the comparison.

        Returns
        -------
        KN
            Right hand side of the comparison.
        """
        raise_if_less_or_equal_to_zero(n=(self.n_s + 1.6))
        return (self.k_1 * (self.n_s / (self.n_s + 1.6)) * ((self.e_cd * self.i_c) / (self.length * M_TO_MM) ** 2)) * N_TO_KN

    @property
    def ratio(self) -> DIMENSIONLESS:
        """Ratio between left hand side and right hand side of the comparison, commonly referred to as unity check."""
        return self.left_hand_side / self.right_hand_side

    def __bool__(self) -> bool:
        """Evaluates the comparison, for more information see the __init__ method."""
        return self.left_hand_side <= self.right_hand_side

    def __str__(self) -> str:
        """Return the result of the comparison."""
        return self.latex().complete

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.18."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"F_{V,Ed} \leq \frac{n_s}{n_s + 1.6} \cdot \frac{\sum E_{cd} \cdot I_c}{L^2}",
            numeric_equation=(
                rf"{self.left_hand_side:.{n}f}"
                rf"\leq \frac{{{self.n_s}}}{{{self.n_s + 1.6}}} \cdot \frac{{\sum {self.e_cd} \cdot {self.i_c}}}{{{self.length**2}}}"
            ),
            comparison_operator_label=r"\rightarrow",
        )
